local PowerStation = {}
local lg = love.graphics


function PowerStation:create( data )
	local entity = Building:create( data )
	
	entity.entityType = ENTITIES.powerStation
	
	-- also store the tile index (avoid to calculate them over and over again)
	entity.xi = data.xi
	entity.yi = data.yi
	
	-- range in full tiles
	entity.explorationRange = 4
	
	entity.image = Images["power_station"]
	entity.rendererLayer = "entities"
	
	-- selection frame image
	entity.imageWidth = 160
	entity.imageHeight = 140
	entity.outlineOffsetY = 40
	local frameImageData = Images["entity_selection_frame_1"]
	entity.selectionFrameImage = frameImageData
	entity.selectionFrameScaleX = entity.imageWidth / frameImageData.image:getWidth()
	entity.selectionFrameScaleY = entity.imageHeight / frameImageData.image:getHeight()
	
	-- blocking info in mini tiles
	-- for now each building can only occupy one full tile
	entity.blocking = {
		{0,0,0,0,0},
		{0,1,1,1,0},
		{0,1,1,1,0},
		{0,1,1,1,0},
		{0,0,0,0,0},
	}
	
	-- workers can deposit resources here
	-- need canDepositEntity and getResourceDepositPosition function
	entity.isResourceDeposit = true
	entity.storedEntities = {}	-- list of lists of entity IDs by entity type
	
	----------------------------------------------------------------------------------------------------
	
	function entity:canDepositEntity( entityType )
		return entityType == ENTITIES.energyCrystal
	end
	
	
	-- absolute world position where workers can deposit resources
	-- important: position must not be blocked!
	function entity:getResourceDepositPosition()
		local x, y = self:getPosition()
		return x - 25, y + 50
	end
	
	
	-- store the entity ID
	-- e.g. an energy crystal
	function entity:storeEntity( entity )
		local eType = entity.entityType
		self.storedEntities[eType] = self.storedEntities[eType] or {}
		table.insert( self.storedEntities[eType], entity.id )
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:setPowerOnState( flag )
		if self.isPowered == flag then
			return
		end
		
		self.isPowered = flag
		-- self:updateImage()
		Beholder.trigger( "OnEntityPowerOnStateChanged", self, self.isPowered )
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:load( id )
		self.id = id
		
		Gameloop:add( self )
		Renderer:add( self, self.rendererLayer )
		
		self.ohwSelection = OhwEntitySelection:create{
			entityId = self.id,
			width = self.imageWidth,
			height = self.imageHeight,
			offsetY = self.outlineOffsetY,
		}
		
		Game:onBuildingCreated( self )
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:tick( dt )
		
		if self.selected then
			self.animationTick = self.animationTick + dt
		end
		
		-- self.timeSinceLastHealthChange = self.timeSinceLastHealthChange + dt
		self.health:update( dt )
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:draw()
		local body = self.body
		local x, y = convertToDimetric( body.x, body.y )
		
		local imgData = self.image
		local ox = imgData.ox or 0
		local oy = imgData.oy or 0
		
		lg.draw( imgData.image, x, y, 0, 1, 1, ox, oy )
		
		-- draw "light"
		-- TODO: do not draw when the building is not operating at the moment
		if true then
			local liSize = 800
			local color = COLORS.buildingLight
			Renderer:drawCircularLight( x, y, liSize, color )
		end
		
		self.ohwSelection:setVisible( self.selected == true )
		
		-- healthbar
		if Game.drawEntityStatusOverhead or self.selected
		or self.health.timeSinceLastHealthChange < 4 then
			self.health:setVisible( true )
			local hb_y_offset = body.z + 120	-- draw above the entity
			self.health:setHealthbarPosition( x, y - hb_y_offset )
		else
			self.health:setVisible( false )
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function entity:destroy()
		Gameloop:remove( self )
		Renderer:remove( self, self.rendererLayer )
		self.ohwSelection:destroy()
		self = nil
	end
	
	
	return entity
end


return PowerStation
