local SwitchButton = {}
local lg = love.graphics

-- idea:
-- similar to a button, but a click iterates through a list of items
--> useful to pick one of X predefined options

local defaultBackgroundColors = { [1] = {0.1, 0.03, 0, 1}, [2] = {0.2, 0.13, 0.1, 1}, [3] = {0.2, 0.13, 0.1, 1}, [4] = {0.2, 0.2, 0.2, 1} }
local defaultTextColors = { [1] = {0.9, 0.75, 0, 1}, [2] = {0.9, 0.75, 0, 1}, [3] = {0.9, 0.45, 0, 1}, [4] = {0.9, 0.75, 0, 1} }
local defaultFrameColors = { [1] = {1, 1, 1, 1}, [2] = {0.9, 0.75, 0, 1}, [3] = {0.9, 0.75, 0, 1}, [4] = {0.4, 0.4, 0.4, 1} }

function SwitchButton:create( data )
	local button = {}
	button.id = data.id
	button.z = 2
	
	button.screenId = data.screenId	-- parent screen (must be!)
	button.parentId = data.parentId	-- parent widget (if any)
	button.canHover = true
	button.canClick = true
	-- button.canDrag = data.canDrag or false
	
	button.callback = data.callback
	button.state = 1
	button.hover = false
	button.active = false
	button.hoverTick = 0
	
	button.enabled = data.enabled ~= false
	button.visible = data.visible ~= false
	
	button.backgroundColors	= Tools.copyTable( defaultBackgroundColors )
	button.textColors	= Tools.copyTable( defaultTextColors )
	button.frameColors = Tools.copyTable( defaultFrameColors )
	
	-- for user interaction (mouse)
	button.body = {}
	button.body.shape = data.shape or "rect"
	button.body.sizeX = data.sizeX or 100
	button.body.sizeY = data.sizeY or 100
	button.body.x = data.x or 0 -- (button.body.sizeX/2)
	button.body.y = data.y or 0 -- (button.body.sizeY/2)
	
	-- optional: auto adjust the button size so that it is quadratic
	-- in that case, the sizeX and sizeY values are upper limits
	if data.autoSizeSquare then
		if button.body.sizeX < button.body.sizeY then
			button.body.sizeY = button.body.sizeX
		elseif button.body.sizeY < button.body.sizeX then
			button.body.sizeX = button.body.sizeY
		end
	end
	
	-- text...
	button.maxFontSize = data.fSize or 40
	button.maxTextWidth = 0.9 * button.body.sizeX
	button.maxTextHeight = 0.9 * button.body.sizeY
	
	----------------------------------------------------------------------------------------------------
	
	-- basically just returns the body (for mouse interaction)
	function button:getShape()
		return self.body
	end
	
	
	function button:setPosition( x, y )
		self.body.x = x or self.body.x
		self.body.y = y or self.body.y
	end
	
	----------------------------------------------------------------------------------------------------
	
	function button:onVisibilityChanged()
		if not self.visible then
			self.hoverTick = 0
			self.tooltipTick = 0
			self.hover = false
			self.active = false
		end
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- click begins (mouse is down; set active in terms of appearance)
	function button:onClickStart()
		self.active = true
		--print( "button start click", self.id )
	end
	
	-- on mouse (left?) click event
	function button:onClickEnd()
		if self.enabled then
			if self.callback then
				self.callback()
			end
			self:_onClicked()
		end
		self.active = false
	end
	
	-- mouse button was released but not while hovering the button ... set inactive and be done
	function button:onClickCancel()
		self.active = false
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- switch to next item
	function button:_onClicked()
		if not self.items then return end
		self.itemIndex = self.itemIndex + 1
		if self.itemIndex > #self.items then
			self.itemIndex = 1
		end
		self:setText( self.items[self.itemIndex] )
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- manage mouseover
	function button:onMouseHoverStart()
		self.hover = true
	end
	
	function button:onMouseHoverEnd()
		self.hover = false
	end
	
	----------------------------------------------------------------------------------------------------
	
	function button:tick(dt)
		if not self.visible then
			return
		end
		
		self.drawTooltip = false
		
		if self.hover then

			self.hoverTick = self.hoverTick + dt
			
			if self.tooltipEnabled and self.hoverTick >= self.tooltipDelay then
				if self.tooltipDuration == -1 or self.tooltipTick < self.tooltipDuration then
					self.drawTooltip = true
					self.tooltipTick = self.tooltipTick + dt
				end
			end
			
		else
		
			self.hoverTick = 0
			self.tooltipTick = 0
			--self.hover = false
			--self.active = false

		end
		
		if not self.enabled then
			self.state = 4
			return
		end
		
		-- from here: execute only if the button is visible and enabled
		
		if self.active then
			self.state = 3
		else
			if self.hover then
				self.state = 2
			else
				self.state = 1
			end
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function button:draw()
		
		if not self.visible then
			return
		end
		
		if self.parentId ~= nil
		and not (GUImgr:isEnabled(self.parentId) and GUImgr:isVisible(self.parentId)) then
			return
		end
		
		local body = self.body
		
		if self.imageData then
			-- print( "draw img")
			local img = self.imageData
			local scaleX = img.scaleX
			local scaleY = img.scaleY
			-- scale the image a bit if the mouse is held down while hovering the button
			if self.state == 3 then
				scaleX = scaleX * 0.9
				scaleY = scaleY * 0.9
			elseif self.state == 4 then
				lg.setColor( COLORS.darkGray )
			end
			-- lg.setScissor( body.x - 0.5 * body.sizeX, body.y - 0.5 * body.sizeY, body.sizeX, body.sizeY )
			lg.draw( img.image, body.x, body.y, 0, scaleX, scaleY, img.offsetX, img.offsetY )
			-- lg.setScissor()
		else
			lg.setColor( self.backgroundColors[self.state] )
			lg.rectangle( "fill", body.x - 0.5 * body.sizeX, body.y - 0.5 * body.sizeY, body.sizeX, body.sizeY, 4, 4 )
		end
		
		lg.setColor( self.frameColors[self.state] )
		lg.rectangle( "line", body.x - 0.5 * body.sizeX, body.y - 0.5 * body.sizeY, body.sizeX, body.sizeY, 4, 4 )
		
		lg.setColor( self.textColors[self.state] )
		
		-- use the defined text width, not the current width of the text object for the horizontal axis
		-- but use the current text height for the y coordinate
		lg.draw( self.text, body.x, body.y, 0, 1, 1, self.maxTextWidth * 0.5, self.text:getHeight() * 0.5 )
		
		
		if self.tooltip then
			self.tooltip.visible = self.drawTooltip
		end
		
		-- love.graphics.setColor(1, 1, 1)
	end
	
	----------------------------------------------------------------------------------------------------
	
	function button:setItems( items, index )
		assert( type(items) == "table" )
		self.items = items
		if index and index <= #items then
			self.itemIndex = index
		else
			self.itemIndex = 1
		end
		self:setText( self.items[self.itemIndex] )
	end
	
	-- TODO: separate displayed text from logic (meaning)?
	-- for now return the text and the index
	function button:getSelectedItem()
		return self.items[self.itemIndex], self.itemIndex
	end
	
	----------------------------------------------------------------------------------------------------
	
	function button:setText( _text, alignMode )
		self.text = self.text or lg.newText( lg.newFont( self.maxFontSize ), "" )
		self.text:setf( _text or "", self.maxTextWidth, alignMode or self.alignMode or "center" )
		self:autoAdjustFontSize( self.text, self.maxTextWidth, self.maxTextHeight )
	end
	
	
	function button:autoAdjustFontSize( text, sizeX, sizeY )
		local fSize = self.maxFontSize or 40
		
		while text:getWidth() >= sizeX
		or text:getHeight() >= sizeY do
			fSize = fSize - 2
			if fSize < 8 then
				print( "button warning: UI element too small for text!", self.id )
				break
			end
			text:setFont( lg.newFont( fSize ) )
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function button:setBackgroundColors( c1, c2, c3, c4 )
		self.backgroundColors[1] = c1 or defaultBackgroundColors[1]
		self.backgroundColors[2] = c2 or defaultBackgroundColors[2]
		self.backgroundColors[3] = c3 or defaultBackgroundColors[3]
		self.backgroundColors[4] = c4 or defaultBackgroundColors[4]
	end
	
	function button:setTextColors( c1, c2, c3, c4 )
		self.textColors[1] = c1 or defaultTextColors[1]
		self.textColors[2] = c2 or defaultTextColors[2]
		self.textColors[3] = c3 or defaultTextColors[3]
		self.textColors[4] = c4 or defaultTextColors[4]
	end
	
	function button:setFrameColors( c1, c2, c3, c4 )
		self.frameColors[1] = c1 or defaultFrameColors[1]
		self.frameColors[2] = c2 or defaultFrameColors[2]
		self.frameColors[3] = c3 or defaultFrameColors[3]
		self.frameColors[4] = c4 or defaultFrameColors[4]
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- function button:setImageData( imageData )
		
		-- if not imageData then
			-- self.imageData = nil
			-- return
		-- end
		
		-- local body = self.body
		
		-- self.imageData = {
			-- image = imageData.image,
			-- scaleX = body.sizeX / imageData.image:getWidth(),
			-- scaleY = body.sizeY / imageData.image:getHeight(),
		-- }
		-- self.imageData.offsetX = 0.5 * body.sizeX / self.imageData.scaleX
		-- self.imageData.offsetY = 0.5 * body.sizeY / self.imageData.scaleY
		
	-- end
	
	----------------------------------------------------------------------------------------------------
	
	function button:initTooltip( tt )
		self.tooltip = Tooltip:create{
			id = self.id .. "_tooltip",
			element = self,
			text = tt.text,
			maxFontSize = tt.maxFontSize,
			color = tt.tooltipColor,
			width = tt.tooltipWidth,
			height = tt.tooltipHeight,
			alignMode = tt.alignMode,
		}
		self.tooltipDelay = tt.delay or 1.0			-- show tt after X seconds
		self.tooltipDuration = tt.duration or 6	-- then show it for up to X seconds (-1 = show forever)
		self.tooltipTick = 0
	end
	
	function button:setTooltipText( text, alignMode )
		if not self.tooltip then
			self:initTooltip{ text = text }
		end
		self.tooltip:setText( text, alignMode )
	end
	
	----------------------------------------------------------------------------------------------------
	
	button:setItems( data.items, data.itemIndex )
	-- button:setText( data.text )
	-- button:setImageData( data.imageData )
	
	-- init tooltip (either exists from the start, or not)
	-- can be turned on and off with tooltipEnabled
	if data.tooltip then
		button.tooltipEnabled = data.tooltip.enabled ~= false
		button:initTooltip( data.tooltip )
	end
	
	Mainloop:add( button )
	GUImgr:addObject( button )
	UIRenderer:add( button )
	
	return button
end


return SwitchButton
