local Game = {}

local lg = love.graphics
local tilesize = TILESIZE
local borderSizeFactor = 0.02	-- for camera movement when mouse is near the screen border

--[[
JOBS
Global Perspective:
- one queue per job type (remove debris, drill wall, ...)
- ordered priorities - can be changed by the player ingame
- if a RR is free (idle) a new job is assigned, search the queue with the highes priority first
- player can manually assign a job to a RR, he will discard his previous job and take on the new task
Rock Raider Perspective:
- has a current job attached
- idle, walking to job or executing job
--]]

--[[
IDEAS FOR MISSION NAMES:
Ärger mit Abraum
Steinreich
Durch das Inferno
Lava-Labyrinth
Sackgasse
Insubordination
Der Ruf der Leere / Call of the void
Es werde Licht / Let there be light
--]]


-- IDEAS:
-- idea for jobs: RR have an idle timer, only consider RR for jobs who have been idle for a second or two
--> a worker who just went to idle state can look for jobs directly next to him
-- carrier/deposit robot where rock raiders can deposit resources until it is full and returns to the base


local GAMESTATE_PLAYING = 1
local GAMESTATE_WON = 2
local GAMESTATE_FAILED = 3


local WallSanityChecks = {
	{-1,-1, 0,-1, -1,0},	-- top left
	{0,-1, 1,-1, 1,0},	-- top right
	{-1,0, -1,1, 0, 1},	-- bottom left
	{0,1, 1,1, 1,0}	-- bottom right
}


function Game:create()
	local game = {}
	
	game.id = "game"
	game.gametime = 0
	game.gamespeed = 1	-- factor -> do not change here; see setGameSpeed
	game.cameraScrollSpeed = 1	-- see guiActionSetScrollSpeed
	
	game.workerJobPriorities = {}
	
	game._debugPaused = false
	
	----------------------------------------------------------------------------------------------------
	
	-- store all jobs in a table so they can be stopped when the player leaves the game
	function game:trackJob( job )
		table.insert( self.simpleJobs, job )
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:open()
		-- GUImgr:setScreen( self.id )
		
		self.gametime = 0
		self._secondTick = 0
		self.playerGlobalResources = {
			[ENTITIES.ore] = 0,
			[ENTITIES.energyCrystal] = 0,
			[ENTITIES.rockSample] = 0,
		}
		self.gamestate = GAMESTATE_PLAYING
		
		-- set to default value for landslide frequency (chance)
		self:setLandslideChance()
		
		-- TODO: air quality decreases over time in most missions
		-- number of units scales with the speed of decay
		self.airQuality = 100
		
		-- allow to cycle through all player controlled rock raider entities
		--> remember the ID of the previous one
		self.lastJumpToRockRaiderId = 0
		
		self:switchToSelectionState( "global" )
		
		-- reset camera scale
		Camera:setScale( 1 )
		
		self:setPaused( false )
		-- self:setGameSpeed( 1.5 )	-- for TESTING
		
		-- healthbar etc.
		-- drawn when a unit takes damage (or gets healed/repaired) or when the entity is selected
		-- or when this variable is true (for all qualified units)
		self.drawEntityStatusOverhead = false
		
		-- reference to minimap (GUI element)
		self.minimap = GUImgr:getObjectById( "gameMinimap" )
		
		-- aboid that the mouse leaves the window area while ingame
		-- love.mouse.setGrabbed( true )
		
		self.onEntityCreatedTriggerId = Beholder.observe( "OnEntityCreated", function(e) game:onEntityCreated(e) end )
		self.onBeforeEntityDestroyedTriggerId = Beholder.observe( "OnBeforeEntityDestroyed", function(e) game:onBeforeEntityDestroyed(e) end )
		self.onEntityPowerStateChangedTriggerId = Beholder.observe( "OnEntityPowerOnStateChanged",
			function(e, state) game:onEntityPowerOnStateChanged(e, state) end )
		self.onEntityDepositedEntityTriggerId = Beholder.observe( "OnEntityDepositedEntity",
			function(carrier, entity, deposit) game:onEntityDepositedEntity(carrier, entity, deposit) end )
		-- self.resolveCarrierHoldEntityTriggerId = Beholder.observe( "EventEverySecond",
			-- function(gametime) game:checkDepositReachableForCarriersHoldingEntity(gametime) end )
		self.onOnEntityHealthDroppedToZeroTriggerId = Beholder.observe( "OnEntityHealthDroppedToZero",
			function(entity) game:onEntityHealthDroppedToZero(entity) end )
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:prepare()
		
		self.selectionState = "global"
		
		self.selectedEntities = {}	-- list of currently selected entities
		
		-- see trackJob
		self.simpleJobs = {}
		
		-- jobs for rock raiders (and or machines?) to perform
		-- actual jobs that wait to be worked on
		self.workerJobs = {}
		
		-- init worker job priorities
		self.workerJobPriorities = {}
		self:setDefaultWorkerJobPriorities()
		self:updatePriorityButtons()
		
		-- important for all jobs: do nothing while game is paused!
		
		-- once per second: manage worker jobs
		self.workerJobAssignmentTriggerId = Beholder.observe( "EventEverySecond", function() game:manageWorkerJobAssignment() end )
		
		-- once per second: update fog of war
		self.updateFogOfWarTriggerId = Beholder.observe( "EventEverySecond", function() game:updateFogOfWar() end )
		
		-- manage everything lava related
		self:trackJob( Simplejob:addJob{
			intervall = 0.1,
			callback = function()
				if GUImgr:getCurrentScreenId() ~= game.id then
					return true
				end
				if game.isPaused then return end
				if game._debugPaused then return end
				
				game:updateLava( 0.1 )
			end,
		} )
		
		
		-- landslides
		self:trackJob( Simplejob:addJob{
			intervall = 3.2,
			callback = function()
				if GUImgr:getCurrentScreenId() ~= game.id then
					return true
				end
				if game.isPaused then return end
				if game._debugPaused then return end
				
				game:manageLandslides()
			end,
		} )
		
		-- minimap (radar)
		--[[
		self:trackJob( Simplejob:addJob{
			intervall = 1,
			callback = function()
				if GUImgr:getCurrentScreenId() ~= game.id then
					return true
				end
				if game.isPaused then return end
				
				game:updateMinimap()
			end,
		} )
		--]]
		
		-- mob spawning
		self.mobSpawner = MobSpawner:create()
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:prepareMap( mapData )
		
		self.terrain = Terrain:create( mapData )
		
		-- set map size
		self.mapSizeX = mapData.sizeX
		self.mapSizeY = mapData.sizeY
		
		self.minimap:setMapSize( self.mapSizeX, self.mapSizeY )
		
		self.energyCrystals = mapData.energyCrystals
		
		-- create walls from map data
		self.wallData = mapData.walls	-- store the walldata
		for yi = 1, mapData.sizeY do
			for xi = 1, mapData.sizeX do
				local wallType = self.wallData[yi][xi]
				if wallType > 0 then
					-- create a wall
					local wall = EntityManager:createEntity(
						wallType,
						-- also store the tile index (avoid to calculate them over and over again)
						{ x = (xi-1) * tilesize, y = (yi-1) * tilesize, xi = xi, yi = yi }
					)
					-- plus, store how many energy crystals can be found inside this wall
					local energyCrystals = self.energyCrystals[yi][xi]
					if energyCrystals > 0 then
						wall.numberOfEnergyCrystals = energyCrystals
					end
				end
			end
		end
		
		-- level generation is complete, now setup the blocking info
		self.blocking = Blocking:create()
		self.blocking:setSize( self.mapSizeX, self.mapSizeY )
		self.blocking:fullBlockingUpdate( self.terrain, self.wallData )
		
		-- fog of war:
		-- terrain tiles and entities are not visible if the area is not discovered yet
		-- terrain is always visible once discovered, entities are only visible if player entities are nearby
		self.fow = FogOfWar:create( self.mapSizeX, self.mapSizeY )
		
		-- after all walls have been created, their images must be changed depending on their neighbours
		-- note: if using "fake walls" for FOW - call this update only AFTER the fog of war has been created!
		local walls = EntityManager:getEntitiesInCategory( ENTITYCATEGORY.wall )
		self:updateWallImageCodes( walls )
		
		-- energy crystals can be inside walls or laying on the ground
		self:createResourcesFromMapData()
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:startCustomMission( map, script )
		
		self:prepare()
		
		self:prepareMap( map )
		
		self:loadScript( script )
		
		self:openMissionBriefingWindow()
		
		-- also update the FOW once at game start
		self:updateFogOfWar()
		
		Logger:log( "Custom game started." )
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- start the game, load a test map
	function game:startTestMap()
		
		self:startLevel( "levels.missions.mission1" )
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:loadScript( script )
		
		self.script = script
		
		-- mission description (briefing)
		GUImgr:getObjectById( "txtGameMissionBriefing" ):setText( self.script.description )
		GUImgr:getObjectById( "btnGameCloseMissionBriefing" ):setText( getString( "game_start_mission" ) )
		
		self:initScriptCallbacks()
		
		-- everything is set up
		self.scriptCallbacks.onGameStart()
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- to be called from the menu
	function game:startLevel( levelPath )
		
		self:prepare()
		
		local script = require( levelPath )
		local mapName = script.mapName
		if not mapName then
			print( "Game ERROR: missing map name!" )
			return
		end
		local mapData = MapLoader:loadMap( mapName )
		
		self:prepareMap( mapData )
		
		self:loadScript( script )
		
		self:openMissionBriefingWindow()
		
		-- also update the FOW once at game start
		self:updateFogOfWar()
		
		Logger:log( "Game started." )
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- called right after game start and from button callback
	function game:openMissionBriefingWindow()
		self:setPaused( true )
		GUImgr:setVisibility( "btnGameOpenMissionBriefing", false )
		GUImgr:setVisibility( "btnGameOpenPauseMenu", false )
		GUImgr:setVisibility( "widGameMissionBriefing", true )
		GUImgr:getObjectById( "txtGameMissionBriefing" ):resetOffset()	-- scroll back to the top
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:initScriptCallbacks()
		
		local script = self.script
		
		self.scriptCallbacks = {}
		
		self.scriptCallbacks.onGameStart = script.onGameStart or function() end
		
		self.scriptCallbacks.onEntityHurtEntity = script.onEntityHurtEntity or function() end
		
		self.scriptCallbacks.onWallDestroyed = script.onWallDestroyed or function() end
		
		self.scriptCallbacks.onWallReinforced = script.onWallReinforced or function() end
		
		self.scriptCallbacks.onBeforeDebrisRemoved = script.onBeforeDebrisRemoved or function() end
		
		self.scriptCallbacks.onBuildingUpgradeStarted = script.onBuildingUpgradeStarted or function() end
		
		self.scriptCallbacks.onBuildingUpgradeCompleted = script.onBuildingUpgradeCompleted or function() end
		
		self.scriptCallbacks.onEntityUpgradeStarted = script.onEntityUpgradeStarted or function() end
		
		self.scriptCallbacks.onEntityUpgradeCompleted = script.onEntityUpgradeCompleted or function() end
		
		-- TODO: add more callbacks
		-- e.g. when entities are created or hurt, when resources are discovered, ...
		--> create dummy callbacks here to avoid plenty of if guards in the game code
		-- TODO: use Beholder for all these events
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- after map start: generate energy crystals that are lying around
	-- for crystals that are inside walls, they will be created when the wall is drilled
	-- note: store the info inside each wall how many crystals there are supposed to be
	-- to make it easier to implement "scanners" that search for resources in walls -> see wall generation above
	function game:createResourcesFromMapData()
		
		local energyCrystals = self.energyCrystals
		local wallData = self.wallData
		local spread = 25
		local random = math.random
		
		for yi = 1, self.mapSizeY do
			for xi = 1, self.mapSizeX do
				local wallType = wallData[yi][xi]
				-- only generate energy crystals if there is no wall
				if wallType == 0 then
					for i = 1, energyCrystals[yi][xi] do
						local energyCrystal = EntityManager:createEntity(
							ENTITIES.energyCrystal,
							{ x = (xi-1) * tilesize + random(-spread,spread), y = (yi-1) * tilesize + random(-spread,spread) }
						)
					end
				end
			end
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- ore has been found (e.g. from mining a wall) at a certain tile
	-- or an energy crystal (mining a wall or simply in a undiscovered cavern area)
	function game:onResourceDiscovered( entity, xi, yi )
		
		-- if not already queued: add a new collect resource job
		local eType = entity.entityType
		if eType == ENTITIES.energyCrystal then
			self:addCollectEnergyCrystalJob( entity )
			self:newNotificationMessage( getString( "game_notification_energy_crystal_discovered" ) )
		elseif eType == ENTITIES.ore then
			self:addCollectOreJob( entity )
			self:newNotificationMessage( getString( "game_notification_ore_discovered" ) )
		end
		
		-- TODO: + cooldown to avoid spam
		-- print( "Game: resource entity found!" )
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- e.g. a RR picked up an energy crystal or a construction fence
	function game:onEntityPickedUpEntity( carrier, entity )
		
		-- print( "game:onEntityPickedUpEntity:", carrier.id, entity.id )
		
		-- a worker picked up a resource entity -> check if there is a bringResourceToBuilding for this entity type
		if entity.entityCategory == ENTITYCATEGORY.resource
		and carrier.checkReadyForJob then
			for jobIndex, job in ipairs( self.workerJobs ) do
				if job.type == "bringResourceToBuilding" and job.resourceType == entity.entityType then
					-- print( "a bringResourceToBuilding job requires this resource type" )
					local worker_x, worker_y = carrier:getPosition()
					local worker_tx, worker_ty = getTileIndex( worker_x, worker_y )
					local constructionSite = EntityManager:getEntityById( job.entityId )
					-- the construction site might have been destroyed
					if constructionSite then
						local constructionSite_x, constructionSite_y = constructionSite:getPosition()
						local target_tx, target_ty = getTileIndex( constructionSite_x, constructionSite_y )
						
						-- print( "check worker..." )
						if carrier:checkReadyForJob( job )
						-- TODO: pathfinding type ground vs flying
						and Pathfinding.getPath( worker_tx, worker_ty, target_tx, target_ty, "ground_coarse" ) then
							-- print( "assign bringResourceToBuilding job" )
							carrier:assignJob( job )
							table.remove( self.workerJobs, jobIndex )
							break	-- only assign one job!
						end
					end
				end
			end
		end
		
	end
	
	
	-- called after the entity has already been dropped
	-- e.g. when a job is canceled
	function game:onEntityDroppedEntity( carrier, entity )
		
		if entity.entityCategory == ENTITYCATEGORY.resource
		and not entity.depositEntity then
			self:addCollectResourceJob( entity )
			
			local eTypeName = GetEntityTypeName( entity.entityType )
			print( "game:onEntityDroppedEntity - add collect resource job", entity.id, eTypeName )
			Logger:log( "game:onEntityDroppedEntity - add collect resource job", entity.id, eTypeName )
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- search a place (entity) to deposit the resource entity for a given worker
	function game:getClosestDeposit( worker, resourceEntity )
		local worker_x, worker_y = worker:getPosition()
		local worker_tx, worker_ty = getTileIndex( worker_x, worker_y )
		local typeToDeposit = resourceEntity.entityType
		
		-- TODO: mobile deposit may be full -> cannot deposit more entities
		local targets = EntityManager:getEntitiesFilter( function(e)
			return e.isResourceDeposit and e:canDepositEntity( typeToDeposit )
		end )
		
		local bestPath
		local bestDeposit
		for _, deposit in pairs( targets ) do
			local x, y = deposit:getPosition()
			local target_tx, target_ty = getTileIndex( x, y )
			local path = Pathfinding.getPath( worker_tx, worker_ty, target_tx, target_ty, "ground_coarse" )
			if path and (not bestPath or #path < bestPath.length) then
				bestPath = {
					path = path,
					length = #path,
					-- targetTileX = target_tx,
					-- targetTileY = target_ty,
				}
				bestDeposit = deposit
			end
		end
		
		return bestDeposit	-- (or nil)
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- a worker brings a resource entity to a construction site
	function game:addResourceEntityToConstructionSite( carriedEntity, constructionSite )
		local resourceType = carriedEntity.entityType
		if not constructionSite:isMissingResource( resourceType ) then
			print( "Game: construction site is not missing resource of type", resourceType, "anymore!" )
			print( "construction site id:", constructionSite.id )
			return
		end
		constructionSite:addResourceProgress( resourceType )
		
		carriedEntity:setDeposit( constructionSite )
		EntityManager:remove( carriedEntity )
		-- print( "game: destroy resource entity for construction site" )
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- called from construction site when all resources are there
	-- the building will be created with 0% construction progress? (TODO)
	function game:replaceConstructionSite( constructionSite, buildingType )
		
		local x, y = constructionSite:getPosition()
		local xi, yi = constructionSite.xi, constructionSite.yi
		local wasSelected = constructionSite.selected == true
		
		print( "game:replaceConstructionSite at", xi, yi )
		print( "...with type", GetEntityTypeName(buildingType) )
		-- print( "construction site id:", constructionSite.id )
		
		Logger:log( "Game:replaceConstructionSite at", xi, yi, "with building type", GetEntityTypeName(buildingType) )
		
		EntityManager:remove( constructionSite )
		
		EntityManager:createEntity(
			buildingType,
			{ x = x, y = y, xi = xi, yi = yi }
		)
		
		-- important: remove all entites from the blocking area!
		-- TODO: find a better way to only execute this for entities with a specific tag?
		if buildingType ~= ENTITIES.buildingArea then
			local blocking = self.blocking
			local entities = EntityManager:getEntitiesFilter( function(e)
				if not e.getPosition
				or not e.setPosition then
					return false
				end
				-- e.moveOutOfBuildingBlocking ?
				local x, y = e:getPosition()
				local exi, eyi = getTileIndex( x, y )
				local mini_xi, mini_yi = getMiniTileIndex( x, y )
				return exi == xi and eyi == yi
					and blocking:isMiniTileBlocked( mini_xi, mini_yi )
			end )
			-- find a close position (mini tile) that is free + (should be?) reachable
			local new_x, new_y
			local mini_xi, mini_yi = getMiniTileIndex( x, y )
			-- local orderedSearchOffsets = { 1, -1, 2, -2, 3, -3, 4, -4, 5, -5 }
			local orderedSearchOffsets = { 1, -1, 2, -2 }
			--> alternative: combine dx and dy offsets to find the closest mini tile
			for i, entity in pairs( entities ) do
				-- look in x direction
				for _, dx in ipairs( orderedSearchOffsets ) do
					if dx ~= 0 and not blocking:isMiniTileBlocked( mini_xi + dx, mini_yi ) then
						new_x, new_y = miniTileIndexToPosition( mini_xi + dx, mini_yi )
						break	-- inner loop, found a position for this entity
					end
				end
				-- look in y direction
				for _, dy in ipairs( orderedSearchOffsets ) do
					if dy ~= 0 and not blocking:isMiniTileBlocked( mini_xi, mini_yi + dy ) then
						new_x, new_y = miniTileIndexToPosition( mini_xi, mini_yi + dy )
						break	-- inner loop, found a position for this entity
					end
				end
				-- update the entity position if a new one was found
				if new_x and new_y then
					Logger:log( "... move entity out of blocking:", entity.id,
						GetEntityTypeName(entity.entityType), new_x, new_y )
					-- print( "... move entity out of blocking:", entity.id,
						-- GetEntityTypeName(entity.entityType), new_x, new_y )
					entity:setPosition( new_x, new_y )
				else
					local ex, ey = entity:getPosition()
					print( "Game:replaceConstructionSite - no position found for entity!", ex, ey )
					Logger:log( "... no position found for entity!", entity.id,
						GetEntityTypeName(entity.entityType), ex, ey )
					-- TODO: remove the entity? move it to the edge of the tile? leave it be?
				end
			end
		end
		
		if wasSelected then
			self:selectTile( xi, yi )
		end
		
	end
			
	----------------------------------------------------------------------------------------------------
	
	-- resources are globally accessible...
	-- when a worker deposits e.g. an ore entity at a tool store
	function game:addToPlayersGlobalResource( resourceType, amount )
		local amount = amount or 1
		self.playerGlobalResources[resourceType] = self.playerGlobalResources[resourceType] + amount
		-- print( "game add player resource;", resourceType, self.playerGlobalResources[resourceType] )
	end
	
	
	function game:getPlayerGlobalResource( resourceType )
		return self.playerGlobalResources[resourceType] or 0
	end
	
	
	-- check if the player has enough resources
	-- cost is a table with the resource type as indices
	-- e.g. hasPlayerEnoughResources{ [ENTITIES.ore] = 8 }
	function game:hasPlayerEnoughResources( cost )
		for k, v in pairs( cost ) do
			if self:getPlayerGlobalResource( k ) < v then
				-- print( "game:hasPlayerEnoughResources - not enough of", k )
				return false
			end
		end
		return true
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- can be called from the mission script
	function game:victory()
		
		if self.gamestate ~= GAMESTATE_PLAYING then
			return fals
		end
		
		self.gamestate = GAMESTATE_WON
		
		Logger:log( "Set game state to won." )
		
		-- TODO: suspend and teleport all units...
		-- ... and wait for all units to be safely teleported
		
		-- go to post game screen
		self:close()
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- iterate over all of the walls to update the code and their image
	function game:updateWallImageCodes( walls )
		for i, wall in pairs( walls ) do
			local xi, yi = wall.xi, wall.yi	-- get the tile index of the wall
			local code = {0,0,0,0,0,0,0,0}	-- representing the 8 neighbours
			local index = 1
			for dy = -1,1 do
				for dx = -1,1 do
					if dx ~= 0 or dy ~= 0 then
						local x = xi + dx
						local y = yi + dy
						if self:isValidTileIndex( x, y ) then
							-- normal variant:
							code[index] = self.wallData[y][x] > 0 and 1 or 0
							
							-- "fake walls" variant:
							-- if self.wallData[y][x] > 0
							-- or not self.fow:isTileDiscovered( x, y ) then
								-- code[index] = 1
							-- else
								-- code[index] = 0
							-- end
						else
							code[index] = 1	-- assume all borders to be walls
						end
						
						index = index + 1
					end
				end
			end
			
			-- print( xi, yi, table.concat(code) )
			wall:updateImageCode( code )
		end
	end
	
	
	function game:isValidTileIndex( xi, yi )
		if xi < 1 or xi > self.mapSizeX
		or yi < 1 or yi > self.mapSizeY then
			return false
		end
		return true
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:updateFogOfWar()
		self.fow:prepareUpdateTick()
		
		-- fetch and apply exploration for all player entities
		local entities = EntityManager:getEntitiesFilter( function(e)
			-- return e.getPosition and e.getExplorationRange
			return e.getPosition and e.getExplorationRange and e.isPlayerControlled
		end )
		
		local exploreTiles = {}	-- tile indices to be explored eventually
		local walls = self.wallData
		local function explore( xi, yi, depth )
			if not self:isValidTileIndex( xi, yi ) then
				return false
			end
			table.insert( exploreTiles, xi )
			table.insert( exploreTiles, yi )
			-- stop when hitting a wall
			if walls[yi][xi] > 0 then
				return false
			end
			-- can we look further?
			depth = depth - 1
			if depth > 0 then
				for dy = -1, 1 do
					for dx = -1, 1 do
						if dx ~= 0 or dy ~= 0 then
							explore( xi+dx, yi+dy, depth )
						end
					end
				end
			end
		end
		
		for _, entity in pairs( entities ) do
			local ex, ey = entity:getPosition()
			local exi, eyi = getTileIndex( ex, ey )
			local range = entity:getExplorationRange()
			
			if entity.explorationIgnoresWalls then
				
				-- discover everything in exploration range around the unit, view not blocked by walls
				-- maybe usful later? available via upgrades? for special units or buildings?
				for dy = -range, range do
					for dx = -range, range do
						local xi = exi + dx
						local yi = eyi + dy
						if Tools.getManhattanDistance2( exi, eyi, xi, yi ) <= range then
							local isNewDiscovery = self.fow:setTileExplored( xi, yi )
							if isNewDiscovery then
								self:onTileDiscovered( xi, yi )
							end
						end
					end
				end
				
			else
			
				exploreTiles = {}
				
				-- explore tile directly
				explore( exi,eyi, range )
				
				-- tiles to be explored are stored in a list: exploreTiles with xi and yi entries
				for i = 1, #exploreTiles, 2 do
					local xi = exploreTiles[i]
					local yi = exploreTiles[i+1]
					if Tools.getManhattanDistance2( exi, eyi, xi, yi ) <= range then
						local isNewDiscovery = self.fow:setTileExplored( xi, yi )
						if isNewDiscovery then
							self:onTileDiscovered( xi, yi )
						end
					end
				end
				
			end
			
		end
		
		-- if we include the "fake walls" to work with the FOW ... update the wall image codes
		-- disabled again because it can result in invalid arrangements (no valid image exists...)
		-- local walls = EntityManager:getEntitiesInCategory( ENTITYCATEGORY.wall )
		-- self:updateWallImageCodes( walls )
	end
	
	
	-- a tile is now explored that was not discovered before
	function game:onTileDiscovered( xi, yi )
		
		-- print( "Game:onTileDiscovered", xi, yi )
		
		-- if there are resources on this tile, inform the player
		local resourceEntities = EntityManager:getEntitiesFilter( function(e)
			if e.entityCategory ~= ENTITYCATEGORY.resource or not e:canBePickedUp() then
				return false
			end
			local ex, ey = e:getPosition()
			local exi, eyi = getTileIndex( ex, ey )
			if exi ~= xi or eyi ~= yi then
				return false
			end
			return true
		end )
		for i, resourceEntity in pairs( resourceEntities ) do
			self:onResourceDiscovered( resourceEntity, xi, yi )
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- lava can flood neighbour tiles
	-- lava deals damage to units on top of it (except they are protected...)
	-- dt is a fixed (but real) value here
	function game:updateLava( dt )
		
		local random = math.random
		local lavaDamage = 10
		local terrain = self.terrain
		local fow = self.fow
		
		-- damage units, TODO: filter
		-- entity must have a health component
		-- entities inside vehicles are not damaged
		-- some entities can be immune to damage from lava
		local entities = EntityManager:getEntitiesFilter( function(e)
			if not e.health then return false end
			if e.isDrivingVehicle and e:isDrivingVehicle() then return false end
			if e.immuneToLavaDamage then return false end
			return true
		end )
		for i, entity in pairs( entities ) do
			local ex, ey = entity:getPosition()
			local exi, eyi = getTileIndex( ex, ey )
			if terrain:getTerrainTypeAtTile( exi, eyi ) == 3 then
				-- hurt entity...
				self:onEntityHurtEntity( nil, entity, lavaDamage )
			end
		end
		
		-- spread to neighbour tile every once in a while
		local function lavaSpread( xi, yi )
			
			local neighbours = {}
			-- direct neighbours only
			local neighbourIndexOffsets = { 0,-1, 1,0, 0,1, -1,0 }
			for i = 1, #neighbourIndexOffsets, 2 do
				local x = xi + neighbourIndexOffsets[i]
				local y = yi + neighbourIndexOffsets[i+1]
				local tile = terrain:getTile( x, y )
				-- only spread to ground tiles and not beneath walls
				-- also do not occupy a tile if there is water next to it
				-- TODO: other conditions that prevent lava from spreading there?
				if tile and tile.type == 1
				and self.wallData[y][x] == 0
				and not self:getBuildingAreaAtTile( x, y )
				and not terrain:hasNeighbourOfType( x, y, 2 ) then
					table.insert( neighbours, tile )
				end
			end
			
			-- per tile chance, TODO: value?
			local chance = 0.05
			for i, tile in pairs( neighbours ) do
				if random() < chance and tile.lavaResistance and tile.lavaResistance > 0 then
					-- little hack for player convenience:
					-- if there is currently a RR whose job it is to repair this exact tile, do not decrease the lavaResistance
					local workers = EntityManager:getEntitiesFilter( function(e)
						if e.entityType ~= ENTITIES.rockRaider then return false end
						local job = e.currentJob
						if not job or job.type ~= "repairErosionDamage" then return false end
						local ex, ey = e:getPosition()
						local xi, yi = getTileIndex( ex, ey )
						if xi ~= tile.xi or yi ~= tile.yi then return false end
						return job.xi == tile.xi and job.yi == tile.yi
					end )
					if #workers == 0 then
						tile.lavaResistance = tile.lavaResistance - 1
						if tile.lavaResistance == 0 then
							terrain:setTerrainTypeAtTile( tile.xi, tile.yi, 3 )
							-- print( "game: set lava at tile", tile.xi, tile.yi )
						end
					else
						Logger:log( "game:lavaSpread - do not spread while a RR is working on it!", tile.xi, tile.yi )
					end
				end
			end
			
		end
		
		for yi = 1, self.mapSizeY do
			for xi = 1, self.mapSizeX do
				local tile = terrain:getTile( xi, yi )
				if tile and tile.type == 3
				and fow:isTileDiscovered( xi, yi ) then
					lavaSpread( xi, yi )
				end
			end
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- update or reset chance for a landslide to occur
	--> per wall chance!
	function game:setLandslideChance( chance )
		self.landslideChance = chance or 0.001
	end
	
	
	function game:getLandslideChance()
		return self.landslideChance
	end
	
	-- called every few(?) seconds
	-- every wall that is not solid and not reinforced can cause a landslide from time to time
	-- if it has a neighbour tile that is free and has normal ground
	function game:manageLandslides()
		local random = math.random
		local chance = self:getLandslideChance()
		local walls = EntityManager:getEntitiesFilter( function(e)
			return e.entityCategory == ENTITYCATEGORY.wall
				and e.allowLandslides
		end )
		for i, wall in pairs( walls ) do
			local canSlide, xi, yi = self:checkLandslideConditions( wall )
			if canSlide then
				if random() < chance then
					self:startLandslide( wall, xi, yi )
				end
			end
		end
	end
	
	
	-- for a given wall entity, check if a landslide can occur right now
	-- that is: the last landslide from this wall lies at least X seconds in the past
	-- and there is a valid tile next to it (no wall, normal ground)
	-- return true and the xi and yi tile index of the free neighbour
	function game:checkLandslideConditions( wall )
		
		local random = math.random
		local minTimeBetweenLandslides = 20	-- time in seconds, TODO?
		
		if wall.isReinforced then
			return false
		end
		
		if not wall.getLastLandslideTime or self.gametime - wall:getLastLandslideTime() < minTimeBetweenLandslides then
			return false
		end
		
		-- free and valid neighbour?
		-- x1,y1, x2,y2, ... for all 4 direct neighbours (not diagonal)
		local neighbourIndexOffsets = { 0,-1, 1,0, 0,1, -1,0 }
		local valid_x, valid_y
		for i = 1, #neighbourIndexOffsets, 2 do
			local x = wall.xi + neighbourIndexOffsets[i]
			local y = wall.yi + neighbourIndexOffsets[i+1]
			if self:isValidTileIndex( x, y ) then
				-- no wall there? ok
				if self.wallData[y][x] == 0
				and self.terrain:getTerrainTypeAtTile( x, y ) == 1 then
					-- ensure both possible neighbours are selected equally
					-- this works because there are always at most 2 possible free neighbour tiles
					if not valid_x or random() < 0.5 then
						valid_x, valid_y = x, y
					end
				end
			end
		end
		
		if valid_x then
			return true, valid_x, valid_y
		else
			return false
		end
	end
	
	-- tile index belongs to a neighbour of the wall
	function game:startLandslide( wall, xi, yi )
		
		local random = math.random
		local x, y = (xi-1) * tilesize, (yi-1) * tilesize
		
		-- entity contains some animation or effect + logic to hurt entities on that tile
		local landslide = EntityManager:createEntity(
			ENTITIES.landslide,
			-- add a small offset so it will be in front of any building at this tile
			{ x = x + 1, y = y + 1, xi = xi, yi = yi, }
		)
		
		-- inform the wall entity (set landslide time)
		wall:setLastLandslideTime( self.gametime )
		
		-- player notification if the area is already discovered
		if self.fow:isTileExplored( xi, yi ) then
			self:newNotificationMessage( getString( "game_notification_landslide" ) )
		end
		
		-- create debris next to the wall, only if there is no debris already!
		local _debris = EntityManager:getEntitiesFilter( function(e)
			return e.entityType == ENTITIES.debris
				and e.xi == xi and e.yi == yi
		end )
		if #_debris > 0 then
			return
		end
		-- ... and no building!
		if self:getBuildingAtTile( xi, yi ) then
			return
		end
		-- if there is a buildingArea -> destroy it
		local buildingArea = self:getBuildingAreaAtTile( xi, yi )
		if buildingArea then
			self:createResourcesForBuildingToDestroy( buildingArea )
			EntityManager:remove( buildingArea )
		end
		
		local debris = EntityManager:createEntity(
			ENTITIES.debris,
			{ x = x, y = y, xi = xi, yi = yi, doNotContainOre = true }
		)
		if self.fow:isTileDiscovered( xi, yi ) then
			self:addClearDebrisJob( debris )
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:updateMinimap()
		
		if not GUImgr:isVisible( "gameMinimap" ) then
			return
		end
		
		local terrainData = {}
		for yi = 1, self.mapSizeY do
			terrainData[yi] = {}
			for xi = 1, self.mapSizeX do
				if self.fow:isTileDiscovered( xi, yi ) then
					terrainData[yi][xi] = self.terrain:getTerrainTypeAtTile( xi, yi )
				end
			end
		end
		
		local walls = EntityManager:getEntitiesFilter( function(e)
			return e.entityCategory == ENTITYCATEGORY.wall
				and self.fow:isTileDiscovered( e.xi, e.yi )
		end )
		
		-- fetch all entities that should be drawn + that are not covered by the FoW
		local entities = EntityManager:getEntitiesFilter( function(e)
			local xi, yi
			if e.entityCategory == ENTITYCATEGORY.building
			or e.entityCategory == ENTITYCATEGORY.buildingArea
			or e.entityCategory == ENTITYCATEGORY.monster
			or e.entityCategory == ENTITYCATEGORY.vehicle
			or e.entityCategory == ENTITYCATEGORY.robot
			or e.entityType == ENTITIES.rockRaider
			or e.entityType == ENTITIES.ore
			or e.entityType == ENTITIES.energyCrystal then
				if e.xi then
					xi, yi = e.xi, e.yi
				else
					-- local ex, ey = e:getPosition()
					xi, yi = getTileIndex( e:getPosition() )
				end
				return self.fow:isTileExplored( xi, yi )
			end
		end )
		
		self.minimap:updateMapData( terrainData, walls, entities )
		
	end
	
	
	function game:minimapJumpTo( x, y )
		Camera:setPosition( x, y )
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- called from script
	-- update the slider accordingly
	function game:setGameSpeed( speed )
		self.gamespeed = speed
		-- print( "game:setGameSpeed", speed )
		Logger:log( "(Script) Set game speed to", speed )
		GUImgr:getObjectById( "sliGameOptionGamespeed" ):setValue( speed )
	end
	
	
	-- called from the UI
	function game:guiActionSetGamespeed( speed )
		self.gamespeed = speed
		Logger:log( "(GUI) Set game speed to", speed )
		-- print( "game:guiActionSetGamespeed", speed )
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:tick( dt )
		
		if GUImgr:getCurrentScreenId() ~= self.id then
			return
		end
		
		GUImgr:getObjectById( "lblGameFps" ):setText( love.timer.getFPS() )
		
		if self.isPaused then return end
		
		if self._debugPaused then return end
		
		-- time that passed in the game world (gamespeed modifier)
		local time = self.gamespeed * dt
		self.gametime = self.gametime + time
		
		-- trigger event every second
		self._secondTick = self._secondTick + dt
		if self._secondTick >= 1 then
			self._secondTick = self._secondTick - 1
			Beholder.trigger( "EventEverySecond", self.gametime )
		end
		
		-- apply gamespeed modifier here
		Gameloop:update( time )
		
		self:updateCameraPosition( dt )
		
		self:updateButtons()
		
		self:updateMinimap()
		
		self:updateValues()
		
	end
	
	function game:updateValues()
		
		-- air quality
		-- bar color changes depending on the current quality
		-- could be changed to be dynamic, but 4 discrete colors should be fine
		local prbGameAirQuality = GUImgr:getObjectById( "prbGameAirQuality" )
		if self.airQuality >= 50 then
			prbGameAirQuality:setBarColor( COLORS.lightGreen )
		elseif self.airQuality >= 25 then
			prbGameAirQuality:setBarColor( COLORS.yellow )
		elseif self.airQuality >= 10 then
			prbGameAirQuality:setBarColor( COLORS.orange )
		else	-- 10 or lower
			prbGameAirQuality:setBarColor( COLORS.red )
		end
		prbGameAirQuality:setProgress( self.airQuality )
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- float value from >0 to 2 or 3
	function game:guiActionSetScrollSpeed( speed )
		self.cameraScrollSpeed = speed
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- camera pan
	function game:updateCameraPosition( dt )
		
		if not WINDOW_HAS_FOCUS then
			return
		end
		
		-- double scroll speed in x direction = good movement when both x AND y movement at the same time
		-- but it feels weird if only x OR y movement occurs
		local cx, cy = 0, 0	-- pan to be applied to the camera
		local mouseX, mouseY = love.mouse.getPosition()
		local width, height = SCREEN.width, SCREEN.height
		
		-- can be moved by mouse...
		-- but only if the mouse does not hover any button(?)
		-- if not GUImgr:isMouseHoveringGuiElement() then
			if mouseX < width * borderSizeFactor then
				-- left
				cx = cx - 2
			elseif mouseX > width * (1 - borderSizeFactor) then
				--right
				cx = cx + 2
			end
			if mouseY < height * borderSizeFactor then
				-- up
				cy = cy - 1
			elseif mouseY > height * (1 - borderSizeFactor) then
				-- down
				cy = cy + 1
			end
		-- end
		
		-- can be moved by keypress...
		if love.keyboard.isDown( Keybinds:getKey("camera_pan_up") ) then
			cy = cy - 1
		elseif love.keyboard.isDown( Keybinds:getKey("camera_pan_down") ) then
			cy = cy + 1
		end
		if love.keyboard.isDown( Keybinds:getKey("camera_pan_left") ) then
			-- cx = cx - 1
			cx = cx - 2
		elseif love.keyboard.isDown( Keybinds:getKey("camera_pan_right") ) then
			-- cx = cx + 1
			cx = cx + 2
		end
		if cx ~= 0 or cy ~= 0 then
			-- make the camera move faster when zoomed in (and slower when zoomed out)
			-- scroll speed can also be adjusted in the pause menu
			local baseSpeed = 200
			local scale = Camera:getScale()
			local cameraSpeed = baseSpeed * self.cameraScrollSpeed / scale
			Camera:move( cx * cameraSpeed * dt, cy * cameraSpeed * dt )
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- an entity is about to hurt another entity
	function game:onEntityHurtEntity( attacker, target, damage )
		
		if not target.health then
			print( "Game:onEntityHurtEntity - target has no health component. Type: ",
				GetEntityTypeName( target.entityType ) )
			return false
		end
		
		local event = {
			entity1 = attacker,
			entity2 = target,
			damage = damage,
		}
		
		target.health:hurt( damage )
		
		self.scriptCallbacks.onEntityHurtEntity( event )
		return event
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- assign jobs to Rock Raiders (and robots?) and check if they are done (?)
	-- gets called every second
	function game:manageWorkerJobAssignment()
		
		local workerJobs = self.workerJobs
		
		-- check all jobs if they are completed
		for i = #workerJobs, 1, -1 do
			local job = workerJobs[i]
			if self:isJobDone( job ) then
				self:onWorkerJobDone( job )
				-- print( "Game: remove completed job", job.type )
				Logger:log( "Game: remove completed job. Type:", job.type )
				table.remove( workerJobs, i )
			end
		end
		-- print( "game #workerJobs:", #workerJobs )
		
		
		-- sort all jobs by 1. priority (job type), 2. time and 3. entityId (to make the sorting function stable)
		local sortFunction = function( a, b )
			if self.workerJobPriorities[a.type] < self.workerJobPriorities[b.type] then
				return true
			elseif self.workerJobPriorities[a.type] > self.workerJobPriorities[b.type] then
				return false
			end
			-- same priority... check time; assign the oldest jobs first
			if a.time < b.time then
				return true
			elseif a.time > b.time then
				return false
			end
			-- finally, sort by associated entity ID
			return a.entityId > b.entityId
		end
		table.sort( workerJobs, sortFunction )
		
		
		-- iterate over all jobs (high priority first) and for all workers who could do the job,
		-- pick the one who has the shortest path; then assign the job
		
		-- local freeWorkers = self:getIdleWorkers()
		local getWorkers = function()
			return EntityManager:getEntitiesFilter( function(e)
				return e.isWorker and e.isPlayerControlled
					and not e.disableAutomaticJobAssigment
					-- workers who are currently selected can still perform their jobs,
					-- but will not be assigned a new one
					--> removed, this is irritating, especially for vehicles (transport truck)
					-- and not e.selected
			end )
		end
		
		local workers = getWorkers()
		for jobIndex, job in ipairs( workerJobs ) do
			
			if job.type == "collectEnergyCrystal" then
				local resourceEntity = EntityManager:getEntityById( job.entityId )
				local resource_x, resource_y = resourceEntity:getPosition()
				local target_tx, target_ty = getTileIndex( resource_x, resource_y )
				local bestPath
				local bestWorker
				
				-- TODO: for both resource types - check if the worker can deposit the resource somewhere?
				
				for i, worker in pairs( workers ) do
					if worker:checkReadyForJob( job ) then
						local worker_x, worker_y = worker:getPosition()
						local worker_tx, worker_ty = getTileIndex( worker_x, worker_y )
						local path = Pathfinding.getPath( worker_tx, worker_ty, target_tx, target_ty, "ground_coarse" )
						-- path may be false! (no path found)
						if path and (not bestPath or #path < bestPath.length) then
							-- this is the first path OR the second path is better than the first...
							bestPath = {
								path = path,
								length = #path,
								targetTileX = target_tx,
								targetTileY = target_ty,
							}
							bestWorker = worker
						end
					end
				end
				
				if bestWorker and bestPath then
					Logger:log( "Game: assign collect energy crystal job to worker " .. bestWorker.id )
					bestWorker:assignJob( job )
					-- remove the job from the queue
					table.remove( self.workerJobs, jobIndex )
					jobAssignedCondition = true
				end
				
			elseif job.type == "collectOre" then
				local resourceEntity = EntityManager:getEntityById( job.entityId )
				local resource_x, resource_y = resourceEntity:getPosition()
				local target_tx, target_ty = getTileIndex( resource_x, resource_y )
				local bestPath
				local bestWorker
				
				for i, worker in pairs( workers ) do
					if worker:checkReadyForJob( job ) then
						local worker_x, worker_y = worker:getPosition()
						local worker_tx, worker_ty = getTileIndex( worker_x, worker_y )
						local path = Pathfinding.getPath( worker_tx, worker_ty, target_tx, target_ty, "ground_coarse" )
						-- path may be false! (no path found)
						if path and (not bestPath or #path < bestPath.length) then
							-- this is the first path OR the second path is better than the first...
							bestPath = {
								path = path,
								length = #path,
								targetTileX = target_tx,
								targetTileY = target_ty,
							}
							bestWorker = worker
						end
					end
				end
				
				if bestWorker and bestPath then
					Logger:log( "Game: assign collect ore job to worker " .. bestWorker.id )
					bestWorker:assignJob( job )
					-- remove the job from the queue
					table.remove( self.workerJobs, jobIndex )
					jobAssignedCondition = true
				end
				
			elseif job.type == "drillWall" then
				
				-- first: check if there is a path from the worker to the target
				-- do not target the wall tile but one neighbour because the wall itself is obviously blocked
				-- check all 4 direct neighbours (not diagonal)
				-- one or two of them should be free
				local wall = EntityManager:getEntityById( job.entityId )
				local xi, yi = wall.xi, wall.yi
				local neighbourIndexOffsets = { 0,-1, 1,0, 0,1, -1,0 }
				local bestPath
				local bestWorker
				
				-- one or two paths possible
				for i = 1, #neighbourIndexOffsets, 2 do
					local target_tx = xi + neighbourIndexOffsets[i]
					local target_ty = yi + neighbourIndexOffsets[i+1]
					-- empty neighbour tile from where a worker could drill this wall
					if self:isValidTileIndex( target_tx, target_ty ) and self.wallData[target_ty][target_tx] == 0 then
						
						for i, worker in pairs( workers ) do
							if worker:checkReadyForJob( job ) then
								local worker_x, worker_y = worker:getPosition()
								local worker_tx, worker_ty = getTileIndex( worker_x, worker_y )
								
								local path = Pathfinding.getPath( worker_tx, worker_ty, target_tx, target_ty, "ground_coarse" )
								-- path may be false! (no path found)
								if path and (not bestPath or #path < bestPath.length) then
									-- this is the first path OR the second path is better than the first...
									bestPath = {
										path = path,
										length = #path,
										targetTileX = target_tx,
										targetTileY = target_ty,
									}
									bestWorker = worker
								end
							end
						end
						
					end
					
				end
				
				if bestWorker and bestPath then
					Logger:log( "Game: assign drill wall job to worker " .. bestWorker.id )
					-- give the info which tile is the best to the worker
					job.targetTileX = bestPath.targetTileX
					job.targetTileY = bestPath.targetTileY
					bestWorker:assignJob( job )
					-- remove the job from the queue
					table.remove( self.workerJobs, jobIndex )
					jobAssignedCondition = true
				end
				
			elseif job.type == "reinforceWall" then
				
				local wall = EntityManager:getEntityById( job.entityId )
				local xi, yi = wall.xi, wall.yi
				local neighbourIndexOffsets = { 0,-1, 1,0, 0,1, -1,0 }
				local bestPath
				local bestWorker
				
				-- one or two paths possible
				for i = 1, #neighbourIndexOffsets, 2 do
					local target_tx = xi + neighbourIndexOffsets[i]
					local target_ty = yi + neighbourIndexOffsets[i+1]
					-- empty neighbour tile from where a worker could reach this wall
					if self:isValidTileIndex( target_tx, target_ty ) and self.wallData[target_ty][target_tx] == 0 then
						
						for i, worker in pairs( workers ) do
							if worker:checkReadyForJob( job ) then
								local worker_x, worker_y = worker:getPosition()
								local worker_tx, worker_ty = getTileIndex( worker_x, worker_y )
								
								local path = Pathfinding.getPath( worker_tx, worker_ty, target_tx, target_ty, "ground_coarse" )
								-- path may be false! (no path found)
								if path and (not bestPath or #path < bestPath.length) then
									-- this is the first path OR the second path is better than the first...
									bestPath = {
										path = path,
										length = #path,
										targetTileX = target_tx,
										targetTileY = target_ty,
									}
									bestWorker = worker
								end
							end
						end
						
					end
					
				end
				
				if bestWorker and bestPath then
					Logger:log( "Game: assign reinforce wall job to worker " .. bestWorker.id )
					-- give the info which tile is the best to the worker
					job.targetTileX = bestPath.targetTileX
					job.targetTileY = bestPath.targetTileY
					bestWorker:assignJob( job )
					-- remove the job from the queue
					table.remove( self.workerJobs, jobIndex )
					jobAssignedCondition = true
				end
				
			elseif job.type == "clearDebris" then
				
				local debris = EntityManager:getEntityById( job.entityId )
				local target_tx, target_ty = debris.xi, debris.yi
				local bestPath
				local bestWorker
				
				for i, worker in pairs( workers ) do
					if worker:checkReadyForJob( job ) then
						local worker_x, worker_y = worker:getPosition()
						local worker_tx, worker_ty = getTileIndex( worker_x, worker_y )
						local path = Pathfinding.getPath( worker_tx, worker_ty, target_tx, target_ty, "ground_coarse" )
						-- path may be false! (no path found)
						if path and (not bestPath or #path < bestPath.length) then
							-- this is the first path OR the second path is better than the first...
							bestPath = {
								path = path,
								length = #path,
								targetTileX = target_tx,
								targetTileY = target_ty,
							}
							bestWorker = worker
						end
					end
				end
				
				if bestWorker and bestPath then
					Logger:log( "Game: assign clear debris job to worker " .. bestWorker.id )
					bestWorker:assignJob( job )
					-- remove the job from the queue
					table.remove( self.workerJobs, jobIndex )
					jobAssignedCondition = true
				end
				
			elseif job.type == "vehicleGetDriver" then
				
				local vehicle = EntityManager:getEntityById( job.entityId )
				local vehicle_x, vehicle_y = vehicle:getPosition()
				local target_tx, target_ty = getTileIndex( vehicle_x, vehicle_y )
				local bestPath
				local bestWorker
				
				for i, worker in pairs( workers ) do
					if worker:checkReadyForJob( job ) then
						local worker_x, worker_y = worker:getPosition()
						local worker_tx, worker_ty = getTileIndex( worker_x, worker_y )
						local path = Pathfinding.getPath( worker_tx, worker_ty, target_tx, target_ty, "ground_coarse" )
						-- path may be false! (no path found)
						if path and (not bestPath or #path < bestPath.length) then
							bestPath = {
								path = path,
								length = #path,
							}
							bestWorker = worker
						end
					end
				end
				
				if bestWorker and bestPath then
					Logger:log( "Game: assign get driver job to worker " .. bestWorker.id )
					bestWorker:assignJob( job )
					-- remove the job from the queue
					table.remove( self.workerJobs, jobIndex )
					jobAssignedCondition = true
				end
				
			elseif job.type == "placeConstructionFence" then
				
				local fence = EntityManager:getEntityById( job.entityId )
				local vehicle_x, vehicle_y = fence:getPosition()
				local target_tx, target_ty = getTileIndex( vehicle_x, vehicle_y )
				local bestPath
				local bestWorker
				
				for i, worker in pairs( workers ) do
					if worker:checkReadyForJob( job ) then
						local worker_x, worker_y = worker:getPosition()
						local worker_tx, worker_ty = getTileIndex( worker_x, worker_y )
						local path = Pathfinding.getPath( worker_tx, worker_ty, target_tx, target_ty, "ground_coarse" )
						-- path may be false! (no path found)
						if path and (not bestPath or #path < bestPath.length) then
							bestPath = {
								path = path,
								length = #path,
							}
							bestWorker = worker
						end
					end
				end
				
				if bestWorker and bestPath then
					Logger:log( "Game: assign place fence job to worker " .. bestWorker.id )
					bestWorker:assignJob( job )
					-- remove the job from the queue
					table.remove( self.workerJobs, jobIndex )
					jobAssignedCondition = true
				end
				
			elseif job.type == "bringResourceToBuilding" then
				
				local constructionSite = EntityManager:getEntityById( job.entityId )
				local constructionSite_x, constructionSite_y = constructionSite:getPosition()
				local target_tx, target_ty = getTileIndex( constructionSite_x, constructionSite_y )
				local bestPath
				local bestWorker
				
				for i, worker in pairs( workers ) do
					if worker:checkReadyForJob( job ) then
						local worker_x, worker_y = worker:getPosition()
						local worker_tx, worker_ty = getTileIndex( worker_x, worker_y )
						local path = Pathfinding.getPath( worker_tx, worker_ty, target_tx, target_ty, "ground_coarse" )
						-- path may be false! (no path found)
						if path and (not bestPath or #path < bestPath.length) then
							bestPath = {
								path = path,
								length = #path,
							}
							bestWorker = worker
						end
					end
				end
				
				if bestWorker and bestPath then
					Logger:log( "Game: assign bringResourceToBuilding job to worker " .. bestWorker.id )
					bestWorker:assignJob( job )
					-- remove the job from the queue
					table.remove( self.workerJobs, jobIndex )
					jobAssignedCondition = true
				end
				
			elseif job.type == "repairErosionDamage" then
				
				local target_tx, target_ty = job.xi, job.yi
				local bestPath
				local bestWorker
				
				for i, worker in pairs( workers ) do
					if worker:checkReadyForJob( job ) then
						local worker_x, worker_y = worker:getPosition()
						local worker_tx, worker_ty = getTileIndex( worker_x, worker_y )
						local path = Pathfinding.getPath( worker_tx, worker_ty, target_tx, target_ty, "ground_coarse" )
						-- path may be false! (no path found)
						if path and (not bestPath or #path < bestPath.length) then
							-- this is the first path OR the second path is better than the first...
							bestPath = {
								path = path,
								length = #path,
								targetTileX = target_tx,
								targetTileY = target_ty,
							}
							bestWorker = worker
						end
					end
				end
				
				if bestWorker and bestPath then
					Logger:log( "Game: assign repairErosionDamage job to worker " .. bestWorker.id )
					bestWorker:assignJob( job )
					-- remove the job from the queue
					table.remove( self.workerJobs, jobIndex )
					jobAssignedCondition = true
				end
				
			elseif job.type == "repairBuilding" then
				
				local building = EntityManager:getEntityById( job.entityId )
				local target_tx, target_ty = building.xi, building.yi
				local bestPath
				local bestWorker	-- DO NOT FORGET to use a local var in each segment here!
				
				for i, worker in pairs( workers ) do
					if worker:checkReadyForJob( job ) then
						local worker_x, worker_y = worker:getPosition()
						local worker_tx, worker_ty = getTileIndex( worker_x, worker_y )
						local path = Pathfinding.getPath( worker_tx, worker_ty, target_tx, target_ty, "ground_coarse" )
						-- path may be false! (no path found)
						if path and (not bestPath or #path < bestPath.length) then
							-- this is the first path OR the second path is better than the first...
							bestPath = {
								path = path,
								length = #path,
								targetTileX = target_tx,
								targetTileY = target_ty,
							}
							bestWorker = worker
						end
					end
				end
				
				if bestWorker and bestPath then
					Logger:log( "Game: assign repairBuilding job to worker " .. bestWorker.id )
					bestWorker:assignJob( job )
					table.remove( self.workerJobs, jobIndex )
					jobAssignedCondition = true
				end
				
				
			end-- job type
			
			-- scan for workers again (one might have a new job already)
			workers = getWorkers()
			
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- cannot directly search a path to a wall tile (because it is blocked)
	-- instead, check all four direct neighbours
	-- return the shortest path or false
	function game:getShortestPathToWall( wall_xi, wall_yi, start_xi, start_yi )
		
		local neighbourIndexOffsets = { 0,-1, 1,0, 0,1, -1,0 }
		local bestPath
		
		for i = 1, #neighbourIndexOffsets, 2 do
			local target_tx = wall_xi + neighbourIndexOffsets[i]
			local target_ty = wall_yi + neighbourIndexOffsets[i+1]
			-- empty neighbour tile from where a worker could drill or reinforce this wall
			if self:isValidTileIndex( target_tx, target_ty ) and self.wallData[target_ty][target_tx] == 0 then
				local path = Pathfinding.getPath( start_xi, start_yi, target_tx, target_ty, "ground_coarse" )
				if path and (not bestPath or #path < bestPath.length) then
					bestPath = {
						path = path,
						length = #path,
						-- targetTileX = target_tx,
						-- targetTileY = target_ty,
					}
					bestWorker = worker
				end
			end
		end
		
		if bestPath then
			return bestPath.path
		else
			return false
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:getIdleWorkers()
		-- check Rock Raiders and vehicles (with drivers)
		local workers = EntityManager:getEntitiesFilter( function(e)
			-- if not e.isWorker or not e.checkReadyForJob then
				-- return false
			-- end
			-- return e:checkReadyForJob()
			return e.task == "idle"
		end )
		return workers
	end
	
	-- deprecated
	function game:tryToAssignJobToWorker( worker )
		
		
		-- print( "no job to assign" )
		return false
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- deprecated?
	function game:guiActionRockRaiderToggleAutomaticJobAssignment()
		local entity = self:getSelectedEntity()
		if not entity
		or entity.entityType ~= ENTITIES.rockRaider then
			return false
		end
		
		entity.disableAutomaticJobAssigment = not entity.disableAutomaticJobAssigment
		
	end
	
	
	-- mode:
	-- 1 = automatically assign new jobs
	-- 2 = manual control, no auto assign
	-- 3 = alarm (combat) mode, no auto assign, can attack monsters (TODO...)
	function game:guiActionSetRockRaiderOpMode( mode )
		local entity = self:getSelectedEntity()
		if not entity
		or entity.entityType ~= ENTITIES.rockRaider then
			return false
		end
		
		entity.disableAutomaticJobAssigment = mode ~= 1
		return true
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- called after job is done
	function game:onWorkerJobDone( workerJob )
		local type = workerJob.type
		-- nothing so far
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- e.g. a RR had a job but was given a different player command (e.g. walk somewhere)
	function game:entityCancelJob( worker, workerJob )
		-- job is done anyway? fine
		if self:isJobDone( workerJob ) then
			return
		end
		
		-- job was canceled anyway (e.g. cancel drill wall)
		if workerJob.canceled then
			return
		end
		
		-- job still needs to be done ... put it back in the right queue
		-- also skip the queuedJobs check because this job has already been queued, was not removed
		-- because it was not done yet, thus it must be queued again (this is no longer the case...?)
		local type = workerJob.type
		if type == "clearDebris" then
			local debris = EntityManager:getEntityById( workerJob.entityId )
			self:addClearDebrisJob( debris, true )
		elseif type == "drillWall" then
			local wall = EntityManager:getEntityById( workerJob.entityId )
			self:addDrillWallJob( wall, true )
		elseif type == "reinforceWall" then
			local wall = EntityManager:getEntityById( workerJob.entityId )
			self:addReinforceWallJob( wall, true )
		elseif type == "collectEnergyCrystal" then
			local energyCrystal = EntityManager:getEntityById( workerJob.entityId )
			self:addCollectEnergyCrystalJob( energyCrystal, true )
		elseif type == "collectOre" then
			local ore = EntityManager:getEntityById( workerJob.entityId )
			self:addCollectOreJob( ore, true )
		elseif type == "bringResourceToBuilding" then
			local constructionSite = EntityManager:getEntityById( workerJob.entityId )
			self:addBringResourceToBuildingJob( constructionSite, workerJob.resourceType )
		elseif type == "placeConstructionFence" then
			local fence = EntityManager:getEntityById( workerJob.entityId )
			self:addPlaceConstructionFenceJob( fence )
		elseif type == "repairErosionDamage" then
			self:addRepairErosionDamageJob( workerJob.xi, workerJob.yi )
		elseif type == "repairBuilding" then
			-- print( "Game:entityCancelJob - queue repairBuilding job again" )
			self:addRepairBuildingJob( EntityManager:getEntityById( workerJob.entityId ) )
		end
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:isJobDone( workerJob )
		
		if not workerJob then
			return true
		end
		
		if workerJob.type == "drillWall" then
			
			return not EntityManager:isValidEntityId( workerJob.entityId )
			
		elseif workerJob.type == "reinforceWall" then
			
			local wall = EntityManager:getEntityById( workerJob.entityId )
			return not wall or wall.isReinforced == true
			
		elseif workerJob.type == "clearDebris" then
			
			return not EntityManager:isValidEntityId( workerJob.entityId )
			
		elseif workerJob.type == "collectEnergyCrystal"
		or workerJob.type == "collectOre"
		or workerJob.type == "collectRockSample" then
			
			local resourceEntity = EntityManager:getEntityById( workerJob.entityId )
			-- resource entity does not exist anymore (because used for building?)
			if not resourceEntity then
				return true
			end
			-- resource entity is now at a deposit
			-- or carried by some entity?
			return resourceEntity.depositEntity ~= nil-- or resourceEntity.carrierEntity ~= nil
			
		elseif workerJob.type == "placeConstructionFence" then
			
			local fence = EntityManager:getEntityById( workerJob.entityId )
			if not fence then
				return true
			end
			-- the fence has been placed at the construction site
			return fence.isPlacedAtConstructionSite == true
			
		elseif workerJob.type == "bringResourceToBuilding" then
			
			local constructionSite = EntityManager:getEntityById( workerJob.entityId )
			if not constructionSite then
				return true
			end
			-- job is done when the construction site is gone
			return false
			
		elseif workerJob.type == "vehicleGetDriver" then
			
			local vehicle = EntityManager:getEntityById( workerJob.entityId )
			if not vehicle then
				return true
			end
			return vehicle:hasDriver()
			
		elseif workerJob.type == "repairErosionDamage" then
			
			local tile = self.terrain:getTile( workerJob.xi, workerJob.yi )
			return tile.type == 3 or tile.lavaResistance == self.terrain.maxLavaResistance
			
		elseif workerJob.type == "repairBuilding" then
			
			-- building is either completely gone or fully repaired
			local building = EntityManager:getEntityById( workerJob.entityId )
			if not building then
				return true
			end
			return building.health:getHealthPercent() == 100
			
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- make sure to not add a new job for the same thing multiple times!
	-- skipQueuedCheck allows to force queue a job again, see entityCancelJob -> not needed anymore?
	
	function game:addDrillWallJob( wall, skipQueuedCheck )
		-- check if the job is already queued
		for i = 1, #self.workerJobs do
			local job = self.workerJobs[i]
			if job.type == "drillWall"
			and job.entityId == wall.id then
				print( "Game: wall is already queued to be drilled", wall.id )
				return false
			end
		end
		
		Logger:log( "Game: new drill wall job " .. wall.id )
		local job = {
			type = "drillWall",
			entityId = wall.id,
			time = self.gametime,
		}
		table.insert( self.workerJobs, job )
		return true
	end
	
	
	function game:addReinforceWallJob( wall )
		for i = 1, #self.workerJobs do
			local job = self.workerJobs[i]
			if job.type == "reinforceWall"
			and job.entityId == wall.id then
				print( "Game: wall is already queued to be reinforced", wall.id )
				return false
			end
		end
		
		Logger:log( "Game: new reinforce wall job " .. wall.id )
		local job = {
			type = "reinforceWall",
			entityId = wall.id,
			time = self.gametime,
		}
		table.insert( self.workerJobs, job )
		return true
	end
	
	
	function game:addClearDebrisJob( debris, skipQueuedCheck )
		for i = 1, #self.workerJobs do
			local job = self.workerJobs[i]
			if job.type == "clearDebris"
			and job.entityId == debris.id then
				print( "Game: debris is already queued to be cleared", debris.id )
				return false
			end
		end
		
		Logger:log( "Game: new clear debris job " .. debris.id )
		local job = {
			type = "clearDebris",
			entityId = debris.id,
			time = self.gametime,
		}
		table.insert( self.workerJobs, job )
		return true
	end
	
	
	-- separate for energy crystal and ore because the player can change the priority individually
	function game:addCollectEnergyCrystalJob( resource )
		for i = 1, #self.workerJobs do
			local job = self.workerJobs[i]
			if job.type == "collectEnergyCrystal"
			and job.entityId == resource.id then
				print( "Game: energy crystal is already queued to be collected", resource.id )
				return false
			end
		end
		
		Logger:log( "Game: new collect energy crystal job " .. resource.id )
		local job = {
			type = "collectEnergyCrystal",
			entityId = resource.id,
			time = self.gametime,
		}
		table.insert( self.workerJobs, job )
		return true
	end
	
	
	function game:addCollectOreJob( resource )
		for i = 1, #self.workerJobs do
			local job = self.workerJobs[i]
			if job.type == "collectOre"
			and job.entityId == resource.id then
				print( "Game: ore is already queued to be collected", resource.id )
				return false
			end
		end
		
		Logger:log( "Game: new collect ore job " .. resource.id )
		local job = {
			type = "collectOre",
			entityId = resource.id,
			time = self.gametime,
		}
		table.insert( self.workerJobs, job )
		return true
	end
	
	-- wrapper to collect ore or energy crystals
	function game:addCollectResourceJob( resource )
		if resource.entityType == ENTITIES.ore then
			return self:addCollectOreJob( resource )
		elseif resource.entityType == ENTITIES.energyCrystal then
			return self:addCollectEnergyCrystalJob( resource )
		end
	end
	
	
	function game:addVehicleGetDriverJob( vehicle )
		for i = 1, #self.workerJobs do
			local job = self.workerJobs[i]
			if job.type == "vehicleGetDriver"
			and job.entityId == vehicle.id then
				print( "Game: vehicle is already queued to be manned", vehicle.id )
				return false
			end
		end
		
		Logger:log( "Game: new get driver job " .. vehicle.id )
		local job = {
			type = "vehicleGetDriver",
			entityId = vehicle.id,
			manuallyAssigned = true,	-- for now this is always true for vehicles
			time = self.gametime,
		}
		table.insert( self.workerJobs, job )
		return true
	end
	
	
	-- for a new building (construction site actually)
	function game:addPlaceConstructionFenceJob( fence )
		Logger:log( "Game: new place construction fence job " .. fence.id )
		local job = {
			type = "placeConstructionFence",
			entityId = fence.id,
			time = self.gametime,
		}
		table.insert( self.workerJobs, job )
		return true
	end
	
	
	-- for each required resource per construction site
	-- no concrete resource entity specified because maybe not exising yet!
	function game:addBringResourceToBuildingJob( constructionSite, resourceType )
		Logger:log( "Game: new bring resource to building job " .. constructionSite.id )
		local job = {
			type = "bringResourceToBuilding",
			entityId = constructionSite.id,
			resourceType = resourceType,
			time = self.gametime,
		}
		table.insert( self.workerJobs, job )
		return true
	end
	
	
	function game:addRepairErosionDamageJob( xi, yi )
		for i = 1, #self.workerJobs do
			local job = self.workerJobs[i]
			if job.type == "repairErosionDamage"
			and job.xi == xi and job.yi == yi then
				print( "Game: erosion damage is already queued to be repaired", xi, yi )
				return false
			end
		end
		
		Logger:log( "Game: new repairErosionDamage job at " .. xi .. ";" .. yi )
		local job = {
			type = "repairErosionDamage",
			xi = xi,
			yi = yi,
			time = self.gametime,
		}
		table.insert( self.workerJobs, job )
		return true
	end
	
	
	function game:addRepairBuildingJob( building )
		for i = 1, #self.workerJobs do
			local job = self.workerJobs[i]
			if job.type == "repairBuilding"
			and job.entityId == building.id then
				print( "Game: building is already queued to be repaired", building.id )
				return false
			end
		end
		
		Logger:log( "Game: new repairBuilding job for building with ID " .. building.id .. " of type " ..
			GetEntityTypeName(building.entityType) )
		local job = {
			type = "repairBuilding",
			entityId = building.id,
			time = self.gametime,
		}
		table.insert( self.workerJobs, job )
		return true
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- look for a mini tile next to a tile (with a wall)
	-- return the mini tile indices which are both inside the neighbour tile
	-- parameters: full tile index of the primary tile and full tile index of the neighbour (TODO: optional?)
	function game:getFreeMiniTileNextToFullTile( xi, yi, nxi, nyi )
		if not self:isValidTileIndex( xi, yi ) then
			print( "Game:getFreeMiniTileNextToFullTile error - invalid tile index!", xi, yi )
			return false
		end
		if not self:isValidTileIndex( nxi, nyi ) then
			print( "Game:getFreeMiniTileNextToFullTile error - invalid neighbour tile index!", nxi, nyi )
			return false
		end
		
		-- if the neighbour tile is to the left of the primary tile (nxi < xi)
		--> check the mini tiles on the right border of the neighbour tile
		if nxi < xi then
			-- tile index * 5 = mini tile index of the last mini tile (bottom right corner)
			--> continue with -2 to get the central mini tile OR -4 to get the first (top left) mini tile
			-- check the central mini tile first, then move towards the corners if needed
			local miniXi = nxi * 5
			-- for miniYi = nyi * 5 - 4, nyi * 5 do
			local y = nyi * 5 - 2	-- central mini tile index
			for _, miniYi in ipairs{ y, y-1, y+1, y-2, y+2 } do
				if not self.blocking:isMiniTileBlocked( miniXi, miniYi ) then
					return miniXi, miniYi
				end
			end
		
		-- neighbour tile is to the right of the primary tile (nxi > xi)
		elseif nxi > xi then
			local miniXi = nxi * 5 - 4
			-- for miniYi = nyi * 5 - 4, nyi * 5 do
			local y = nyi * 5 - 2	-- central mini tile index
			for _, miniYi in ipairs{ y, y-1, y+1, y-2, y+2 } do
				if not self.blocking:isMiniTileBlocked( miniXi, miniYi ) then
					return miniXi, miniYi
				end
			end
		
		-- neighbour tile is above the primary tile (nyi < yi)
		elseif nyi < yi then
			local x = nxi * 5 - 2	-- central mini tile index
			local miniYi = nyi * 5
			-- for miniXi = nxi * 5 - 4, nxi * 5 do
			for _, miniXi in ipairs{ x, x-1, x+1, x-2, x+2 } do
				if not self.blocking:isMiniTileBlocked( miniXi, miniYi ) then
					return miniXi, miniYi
				end
			end
		
		-- neighbour tile is below the primary tile (nyi > yi)
		elseif nyi > yi then
			local x = nxi * 5 - 2	-- central mini tile index
			-- for miniXi = nxi * 5 - 4, nxi * 5 do
			local miniYi = nyi * 5 - 4
			for _, miniXi in ipairs{ x, x-1, x+1, x-2, x+2 } do
				if not self.blocking:isMiniTileBlocked( miniXi, miniYi ) then
					return miniXi, miniYi
				end
			end
		
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:onKeyPressed( key, scancode )
		
		local action = Keybinds:getAction( key )
		if not action then return end
		
		if action == "toggle_fps_count" then
			GUImgr:toggleVisibility( "lblGameFps" )
			
		elseif action == "toggle_entity_overhead" then
			self.drawEntityStatusOverhead = not self.drawEntityStatusOverhead
			
		elseif action == "debug_pause" then
			-- DEBUG ONLY
			self._debugPaused = not self._debugPaused
			
		elseif action == "debug_toggle_fow" then
			-- DEBUG ONLY
			if self.fow then
				print( "debug: toggle FoW" )
				self.fow.debugExploreAll = not self.fow.debugExploreAll
			end
			
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:onMousePressed( _x, _y, button )
		
		if GUImgr:getCurrentScreenId() ~= self.id then
			return
		end
		
		if self.isPaused then return end
		if self._debugPaused then return end
		
		-- ignore click if there is any UI element in front...
		if GUImgr:isMouseHoveringGuiElement() then
			return
		end
		
		-- print( "game: onMousePressed", _x, _y )
		local wx, wy = Camera:getTransform():inverseTransformPoint( _x, _y )
		-- print( "world:", wx, wy )
		local x, y = convertTo2D( wx, wy )	-- naming for conversion seems confusing...
		-- print( "world:", x, y )
		local xi, yi = getTileIndex( x, y )
		local mini_xi, mini_yi = getMiniTileIndex( x, y )
		-- print( xi, yi )
		
		-- if selection state...
		-- select the tile or the building / wall that is placed there
		if button == 1 then	-- left click
			self:clearEntitySelection()
			
			-- selecting an entity such as Rock Raiders requires to check for 2d coordinates, not world coordinates!
			local entityAtMousePosition = self:getEntityAtScreenPosition( _x, _y )
			if entityAtMousePosition and entityAtMousePosition.isPlayerControlled then
				self:selectEntity( entityAtMousePosition )
				-- local ex, ey = entityAtMousePosition:getPosition()
				-- local start_xi, start_yi = getMiniTileIndex( ex, ey )
				-- print( "start:", start_xi, start_yi )
				return
			end
			
			-- select wall by screen coordinates
			-- the approach is good ... but my implementation is faulty
			-- two things (probably) cause trouble:
			-- the scaling (zoomed in often results in pixel index out of bounds error)
			-- the image size and using half the outline size to transform the point inside isPointIn2dHitbox
			-- TODO ...
			-- local wallAtMousePosition = self:getWallAtScreenPosition( _x, _y )
			-- if wallAtMousePosition then
				-- self:selectEntity( wallAtMousePosition )
				-- return
			-- end
			
			-- this has been moved behind the entity and wall selection
			-- to work with wall selection
			-- but might be problematic with not discovered entities?
			if not self.fow:isTileDiscovered( xi, yi ) then
				return
			end
			
			if self:isValidTileIndex( xi, yi ) then
				self:selectTile( xi, yi )
			-- else
				-- self:selectTile()
			end
			
		elseif button == 2 then	-- right click
			
			-- do nothing when clicking inside an undiscovered area
			if not self.fow:isTileDiscovered( xi, yi ) then
				return
			end
			
			-- if RR selected -> perform action (walk, drill wall, ...)
			local selectedEntities = self:getSelectedEntities()
			local entityAtMousePosition = self:getEntityAtScreenPosition( _x, _y )
			
			-- click on ground = move there, wall = add drill job, ...
			for i, entity in pairs( selectedEntities ) do
				
				-- interaction with selected entity? (from 2d hitbox)
				-- TODO ...
				
				if entity.isWorker and entity:canIssueMoveCommand() then
					local ex, ey = entity:getPosition()
					local worker_tx, worker_ty = getTileIndex( ex, ey )
					
					-- drill wall
					local walls = EntityManager:getEntitiesFilter( function(e)
						return e.entityCategory == ENTITYCATEGORY.wall
							and e.canBeDrilled
							and e.xi == xi and e.yi == yi
					end )
					if walls[1] and entity:canPerformJobType( "drillWall" ) then
						local wall = walls[1]
						-- has a path?
						local path = self:getShortestPathToWall( wall.xi, wall.yi, worker_tx, worker_ty )
						if not path then
							print( "Game info - no path to wall" )
							return
						end
						-- let the worker drill the wall
						local job = {
							type = "drillWall",
							entityId = wall.id,
							time = self.gametime,
							targetTileX = path[1].x,	-- required for the worker... because there are multiple potential neighbour tiles to go
							targetTileY = path[1].y,
							-- important: mark this job as created from player command
							--> it does not go into the global queue, it will not be listed again if the job is canceled
							manuallyAssigned = true,
						}
						entity:assignJob( job )
						return
					end
					
					-- collect resource entity
					if entityAtMousePosition
					and entity:canPerformJobType( "collectEnergyCrystal" )
					and entityAtMousePosition.entityType == ENTITIES.energyCrystal
					and not entityAtMousePosition.carrierEntity then
						if Pathfinding.getPath( xi, yi, worker_tx, worker_ty, "ground_coarse" ) then
							local job = {
								type = "collectEnergyCrystal",
								entityId = entityAtMousePosition.id,
								time = self.gametime,
								manuallyAssigned = true,
							}
							entity:assignJob( job )
							return
						end
					end
					
					if entityAtMousePosition
					and entity:canPerformJobType( "collectOre" )
					and entityAtMousePosition.entityType == ENTITIES.ore
					and not entityAtMousePosition.carrierEntity then
						if Pathfinding.getPath( xi, yi, worker_tx, worker_ty, "ground_coarse" ) then
							local job = {
								type = "collectOre",
								entityId = entityAtMousePosition.id,
								time = self.gametime,
								manuallyAssigned = true,
							}
							entity:assignJob( job )
							return
						end
					end
					
					if entityAtMousePosition
					and entity:canPerformJobType( "collectRockSample" )
					and entityAtMousePosition.entityType == ENTITIES.rockSample
					and not entityAtMousePosition.carrierEntity then
						if Pathfinding.getPath( xi, yi, worker_tx, worker_ty, "ground_coarse" ) then
							local job = {
								type = "collectRockSample",
								entityId = entityAtMousePosition.id,
								time = self.gametime,
								manuallyAssigned = true,
							}
							entity:assignJob( job )
							return
						end
					end
					
					-- drive vehicle
					if entityAtMousePosition
					and entity.entityType == ENTITIES.rockRaider
					and entityAtMousePosition.canBeManned and entityAtMousePosition:canBeManned() then
					-- and entityAtMousePosition.entityCategory == ENTITYCATEGORY.vehicle then
						if Pathfinding.getPath( xi, yi, worker_tx, worker_ty, "ground_coarse" ) then
							local job = {
								type = "vehicleGetDriver",
								entityId = entityAtMousePosition.id,
								time = self.gametime,
								-- important: mark this job as created from player command
								--> it does not go into the global queue, it will not be listed again if the job is canceled
								manuallyAssigned = true,
							}
							entity:assignJob( job )
							return
						end
					end
					
					-- clear debris
					local _debris = EntityManager:getEntitiesFilter( function(e)
						return e.entityType == ENTITIES.debris
							and e.xi == xi and e.yi == yi
					end )
					if _debris[1] and entity:canPerformJobType( "clearDebris" ) then
						local debris = _debris[1]
						local path = Pathfinding.getPath( xi, yi, worker_tx, worker_ty, "ground_coarse" )
						if not path then
							print( "Game info - no path to debris" )
							return
						end
						local job = {
							type = "clearDebris",
							entityId = debris.id,
							time = self.gametime,
							-- important: mark this job as created from player command
							--> it does not go into the global queue, it will not be listed again if the job is canceled
							manuallyAssigned = true,
						}
						entity:assignJob( job )
						return
					end
					
				end
				
				-- move
				if entity.moveToDestination then
					if entity:canIssueMoveCommand() then
						entity:moveToDestination( mini_xi, mini_yi, x, y )
						return
					else
						-- entity cannot move right now - why?
						-- maybe it is best to just ignore it
						-- example case: RR follows his path, suddenly a building blocks the path,
						-- but the RR keeps following this path
						--> the entity might have a good reason to not be able to move now
						--> e.g. a RR that just arrives at a teleport pad is inside the building blocking
						--> but he should NOT be move out of there by force!
					end
				end
				
			end
			
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:onWheelMoved( dx, dy )
		
		if GUImgr:isMouseHoveringGuiElement() or self.isPaused then
			return
		end
		
		-- camera zoom
		if dy < 0 then
			-- moved down
			Camera:zoomOut()
		elseif dy > 0 then
			-- moved up
			Camera:zoomIn()
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- note: while this works fine for selecting units, it has one drawback:
	-- when right clicking on a vehicle while a RR is selected and standing directly next to the vehicle,
	-- the RR will be returned again by this function (so he will walk instead of getting into the vehicle)
	
	-- there might be multiple entities at this position...
	function game:getEntityAtScreenPosition( _x, _y )
		local point = { x = _x, y = _y }
		local filter = function(e)
			if not e.get2dHitbox then return false end
			local hitbox = e:get2dHitbox()
			return Tools.isPointInsideShape( point, hitbox )
			-- if e.get2dHitbox then
				-- return Tools.isPointInsideShape( point, e:get2dHitbox() )
			-- elseif e.isPointIn2dHitbox then
				-- return e:isPointIn2dHitbox( point )
			-- end
		end
		local entities = EntityManager:getEntitiesFilter( filter )
		-- print( "Game:getEntityAtScreenPosition - #entities", #entities )
		if #entities == 0 then
			return false
		end
		-- sort entities by their coordinate same as the renderer
		-- important: the first entity in the list will be the furthest in the back!
		local sortFunction = Renderer.sortFunction
		table.sort( entities, sortFunction )
		
		-- always try to pick Rock Raiders first
		-- for i, entity in pairs( entities ) do
		for i = #entities, 1, -1 do
			local entity = entities[i]
			if entity.entityType == ENTITIES.rockRaider then
				return entity
			end
		end
		-- otherwise, return the last hit (last because of renderer sort order)
		return entities[#entities]
	end
	
	-- similar version: only select a player controlled entity!
	-- there might still be multiple entities at this position...
	function game:getPlayerControlledEntityAtScreenPosition( _x, _y )
		local point = { x = _x, y = _y }
		local filter = function(e)
			if not e.get2dHitbox then return false end
			if not e.isPlayerControlled then return false end
			local hitbox = e:get2dHitbox()
			return Tools.isPointInsideShape( point, hitbox )
		end
		local entities = EntityManager:getEntitiesFilter( filter )
		-- print( "Game:getPlayerControlledEntityAtScreenPosition - #entities", #entities )
		if #entities == 0 then
			return false
		end
		-- always try to pick Rock Raiders first
		for i, entity in pairs( entities ) do
			if entity.entityType == ENTITIES.rockRaider then
				return entity
			end
		end
		-- otherwise, return the first hit
		return entities[1]
	end
	
	
	function game:getWallAtScreenPosition( _x, _y )
		local point = { x = _x, y = _y }
		local filter = function(e)
			if e.entityCategory == ENTITYCATEGORY.wall
			and e.entityType ~= ENTITIES.solidRock
			and e.isPointIn2dHitbox then
				return e:isPointIn2dHitbox( point )
			end
			return false
		end
		local entities = EntityManager:getEntitiesFilter( filter )
		print( "Game:getWallAtScreenPosition - #entities", #entities )
		if #entities == 0 then
			return false
		end
		-- sort entities by their coordinate same as the renderer
		-- important: the first entity in the list will be the furthest in the back!
		local sortFunction = Renderer.sortFunction
		table.sort( entities, sortFunction )
		
		-- otherwise, return the last hit (last because of renderer sort order)
		return entities[#entities]
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:getSelectedEntities()
		return self.selectedEntities
	end
	
	-- return the first of all selected entities
	function game:getSelectedEntity()
		return self.selectedEntities[1]
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:selectTile( xi, yi )
		
		if self.terrain then
			self.terrain:selectTile( xi, yi )
		end
		
		if not xi then
			self.selectedTile = nil
			self:switchToSelectionState( "global" )
			return false
		end
		
		self.selectedTile = { xi = xi, yi = yi }
		
		self:switchToSelectionState( "tile" )
		
		-- wall
		local wallSelected = false
		local walls = EntityManager:getEntitiesInCategory( ENTITYCATEGORY.wall )
		for i, wall in pairs( walls ) do
			if wall.xi == xi and wall.yi == yi then
				if wall.entityType == ENTITIES.solidRock then	-- do not select solid rock...
					self:selectTile()
					return false
				end
				-- self:switchToSelectionState( "wall" )
				self:selectEntity( wall )
				-- print( "Game: select wall at", xi, yi )
				wallSelected = true
				break
			end
		end
		
		-- building
		if not wallSelected then
			local buildingSelected = false
			local building = self:getBuildingAtTile( xi, yi )
			if building then
				-- self:switchToSelectionState( "building" )
				self:selectEntity( building )
				buildingSelected = true
			end
			
			-- building area?
			if not buildingSelected then
				local buildingAreaSelected = false
				local buildings = EntityManager:getEntitiesInCategory( ENTITYCATEGORY.buildingArea )
				for i, building in pairs( buildings ) do
					if building.xi == xi and building.yi == yi then
					-- if building.xi == xi and building.yi == yi
					-- and building.entityType ~= ENTITIES.buildingArea then
						-- self:switchToSelectionState( "tile" )
						self:selectEntity( building )
						buildingAreaSelected = true
						break
					end
				end
			
				-- debris?
				if not buildingAreaSelected then
					local debris = EntityManager:getEntitiesInCategory( ENTITYCATEGORY.debris )
					for i, d in pairs( debris ) do
						if d.xi == xi and d.yi == yi then
							-- self:switchToSelectionState( "tile" )
							self:selectEntity( d )
							break
						end
					end
				end
			end
		end
		
	end
	
	
	function game:reselectCurrentTile()
		if not self.selectedTile then
			return
		end
		self:selectTile( self.selectedTile.xi, self.selectedTile.yi )
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:clearEntitySelection()
		if self.selectedEntity then
			if self.selectedEntity.setSelected then
				self.selectedEntity:setSelected( false )
			end
		end
		self.selectedEntity = nil
		self.selectedEntities = {}
		self:selectTile()
		self:onEntitySelectionChanged()
	end
	
	
	-- player selected a wall, building or mobile entity
	function game:selectEntity( entity )
		-- print( "game:selectEntity of type", GetEntityTypeName(entity.entityType) )
		
		-- if the entity is currently driving a vehicle, select the vehicle instead
		if entity.getVehicle then
			local vehicle = entity:getVehicle()
			if vehicle and vehicle.setSelected then
				return self:selectEntity( vehicle )
			end
		end
		
		self:clearEntitySelection()
		
		if entity.setSelected then
			entity:setSelected( true )
		end
		self.selectedEntity = entity
		self.selectedEntities = { entity }
		-- print( "Game: selectEntity", entity.id, entity.entityType )
		-- self:switchToSelectionState( "entity" )
		
		-- switch selection state
		if entity.entityCategory == ENTITYCATEGORY.wall then
			self:switchToSelectionState( "wall" )
		elseif entity.entityCategory == ENTITYCATEGORY.building then
			self:switchToSelectionState( "building" )
		elseif entity.entityCategory == ENTITYCATEGORY.buildingArea
		or entity.entityCategory == ENTITYCATEGORY.debris then
			self:switchToSelectionState( "tile" )
		else
			self:switchToSelectionState( "entity" )
		end
		
		self:onEntitySelectionChanged()
	end
	
	
	-- called whenever a new entity was added to the selection or removed from it
	function game:onEntitySelectionChanged()
		
		-- print( "game:onEntitySelectionChanged", self.selectionState )
		
		-- hide all first?
		GUImgr:setVisibility( "widGameBuildingMenu", false )
		GUImgr:setVisibility( "widGameVehicleFactoryNewVehicleMenu", false )
		GUImgr:setVisibility( "widGameRobotBayNewRobotMenu", false )
		GUImgr:setVisibility( "widGameRockRaiderOperationModeMenu", false )
		GUImgr:setVisibility( "btnGameActionTeleportRockRaiderToBuilding", false )
		GUImgr:setVisibility( "btnGameActionUpgradeBuilding", false )
		GUImgr:setVisibility( "btnGameActionRepairBuilding", false )
		GUImgr:setVisibility( "btnGameActionExpellUnit", false )
		GUImgr:setVisibility( "btnGameActionEntityCancelCurrentJob", false )
		-- GUImgr:setVisibility( "btnGameActionEntityDropCarriedEntities", false )
		-- GUImgr:setVisibility( "btnGameActionRockRaiderToggleAutomaticJobAssignment", false )
		GUImgr:setVisibility( "btnGameToggleRockRaiderOpModeMenu", false )
		GUImgr:setVisibility( "btnGameActionVehicleGetDriver", false )
		GUImgr:setVisibility( "btnGameActionVehicleDriverLeave", false )
		GUImgr:setVisibility( "btnGameVehicleFactoryToggleNewVehicleMenu", false )
		GUImgr:setVisibility( "btnGameRobotBayToggleNewRobotMenu", false )
		GUImgr:setVisibility( "widGameWorkerJobPriorities", false )
		GUImgr:setVisibility( "btnGameToggleUpgradeEntityMenu", false )
		GUImgr:setVisibility( "widGameUpgradeEntityMenu", false )
		
		GUImgr:setEnabledFlag( "btnGameActionRemoveDebris", false )
		GUImgr:setEnabledFlag( "btnGameToggleBuildingMenu", false )
		
		
		-- TODO: show widget...
		local selectedEntities = self:getSelectedEntities()
		if #selectedEntities == 0 then
			
			-- nothing selected, reset?
			
			return
		end
		
		local entity = selectedEntities[1]
		local entityType = entity.entityType
		local entityCategory = entity.entityCategory
		
		if entityCategory == ENTITYCATEGORY.vehicle then
			GUImgr:setVisibility( "btnGameActionVehicleGetDriver", true )
			GUImgr:setVisibility( "btnGameActionVehicleDriverLeave", true )
			GUImgr:setVisibility( "btnGameActionExpellUnit", true )
			GUImgr:setVisibility( "btnGameToggleUpgradeEntityMenu", true )
		end
		
		if entityType == ENTITIES.rockRaider then
			-- GUImgr:setVisibility( "btnGameActionRockRaiderToggleAutomaticJobAssignment", true )
			GUImgr:setVisibility( "btnGameToggleRockRaiderOpModeMenu", true )
		end
		
		-- GUImgr:setEnabledFlag( "btnGameActionRemoveDebris", entityType == ENTITIES.debris )
		
		if entityCategory == ENTITYCATEGORY.building then
			
			if entityType ~= ENTITIES.constructionSite then
				GUImgr:setVisibility( "btnGameActionUpgradeBuilding", true )
				GUImgr:setVisibility( "btnGameActionRepairBuilding", true )
			end
		end
		
		if entityType == ENTITIES.teleportPad then
			GUImgr:setVisibility( "btnGameActionTeleportRockRaiderToBuilding", true )
		elseif entityType == ENTITIES.vehicleFactory then
			GUImgr:setVisibility( "btnGameVehicleFactoryToggleNewVehicleMenu", true )
		elseif entityType == ENTITIES.robotBay then
			GUImgr:setVisibility( "btnGameRobotBayToggleNewRobotMenu", true )
		elseif entityType == ENTITIES.transportTruck then
			-- GUImgr:setVisibility( "btnGameActionEntityDropCarriedEntities", true )
		end
		
		if entity.isWorker then
			GUImgr:setVisibility( "btnGameActionEntityCancelCurrentJob", true )
			GUImgr:setVisibility( "btnGameActionExpellUnit", true )
		end
		
		if entityCategory == ENTITYCATEGORY.robot then
			GUImgr:setVisibility( "btnGameActionExpellUnit", true )
			GUImgr:setVisibility( "btnGameToggleUpgradeEntityMenu", true )
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- adding a building = add its blocking data
	-- removing a building = simply use blocking:singleTileUpdate
	function game:onBuildingCreated( entity )
		if not entity.getBlocking
		or not entity.xi or not entity.yi then
			print( "Game:onBuildingCreated error - invalid tile index or no blocking", entity.xi, entity.yi )
			return false
		end
		
		local blocking = entity:getBlocking()
		self.blocking:addEntityBlockingAtTile( entity.xi, entity.yi, blocking )
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- entity (actor) may be a Rock Raider
	function game:entityRemovesMassFromDebris( entity, debris )
		if debris.removeMass then
			debris:removeMass()
		else
			print( "Game:entityRemovesMassFromDebris warning - invalid debris entity" )
		end
	end
	
	-- entity (actor) may be a loader dozer
	function game:entityRemovesAllMassFromDebris( entity, debris )
		if debris.removeLastMass then
			debris:removeLastMass()
		else
			print( "Game:entityRemovesAllMassFromDebris warning - invalid debris entity" )
		end
	end
	
	-- last mass removed from debris
	--> right before the debris is gone
	--> reveal (create) ore if the debris does originate from a drilled wall
	function game:onBeforeDebrisRemoved( debris )
		if debris.doNotContainOre then
			return
		end
		
		local x, y = debris:getPosition()
		local xi, yi = getTileIndex( x, y )
		-->use a sequence instead to make it look random but have a "good" distribution?
		local random = math.random
		local amount = random( 1, 4 )	-- TODO
		local spread = 50
		for i = 1, amount do
			local ore = EntityManager:createEntity(
				ENTITIES.ore,
				{ x = x + random(-spread,spread), y = y + random(-spread,spread) }
			)
			self:onResourceDiscovered( ore, xi, yi )
		end
		
		-- script callback
		self.scriptCallbacks.onBeforeDebrisRemoved( debris, xi, yi )
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- entity may be a Rock Raider (or a robot...)
	-- who is currently drilling a wall entity
	-- works over time, so need dt!
	function game:entityDrillWall( entity, wall, dt )
		
		-- the driller has some value that resembles his efficiency
		-- the wall has a value that states how hard it is to drill it
		if not wall.drillEndurance
		or not entity.drillWallEfficiency then
			print( "Game warning: cannot drill wall (missing properties)" )
			return false
		end
		
		local newEndurance = wall.drillEndurance - entity.drillWallEfficiency * dt
		if newEndurance > 0 then
			wall:setRemainingDrillEndurance( newEndurance )
		else
			self:removeWall( wall )
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- a Rock Raider (first param) is done reinforcing a wall
	function game:entityReinforceWall( entity, wall )
		
		-- if not wall.entityCategory == ENTITYCATEGORY.wall
		if not wall.reinforce then
			print( "game error: try to reinforce a non-wall entity" )
			return false
		end
		
		-- a reinforced wall will not cause a landslide anymore
		-- wall.isReinforced = true
		wall:reinforce()
		
		-- script callback: RR reinforced a wall
		self.scriptCallbacks.onWallReinforced( wall )
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- tile is what game.terrain:getTile returns
	function game:entityRepairErosionDamage( entity, tile, amount )
		tile.lavaResistance = tile.lavaResistance + amount
		if tile.lavaResistance > self.terrain.maxLavaResistance then
			tile.lavaResistance = self.terrain.maxLavaResistance
			-- print( "repairing erosion damage is done." )
		end
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- amount = flat health value
	-- called every ? seconds (defined in the script of the actor entity)
	function game:entityRepairBuilding( entity, building, amount )
		-- print( "game:entityRepairBuilding" )
		-- print(  entity, building, amount )
		
		local amount = amount or 1
		if not building or not building.health then
			Logger:log( "game:entityRepairBuilding ERROR - invalid building/missing health component" )
			return false
		end
		local health = building.health:getHealth()
		building.health:setHealth( health + amount )
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:close()
		
		-- remove all listeners
		Beholder.stopObserving( self.onEntityCreatedTriggerId )
		Beholder.stopObserving( self.onBeforeEntityDestroyedTriggerId )
		Beholder.stopObserving( self.onEntityPowerStateChangedTriggerId )
		Beholder.stopObserving( self.onEntityDepositedEntityTriggerId )
		-- Beholder.stopObserving( self.resolveCarrierHoldEntityTriggerId )
		Beholder.stopObserving( self.onOnEntityHealthDroppedToZeroTriggerId )
		Beholder.stopObserving( self.workerJobAssignmentTriggerId )
		Beholder.stopObserving( self.updateFogOfWarTriggerId )
		
		EntityManager:destroyAllEntities()
		if self.terrain then
			self.terrain:destroy()
		end
		
		self.mobSpawner:destroy()
		
		-- remove all jobs
		-- see trackJob
		for i, job in pairs( self.simpleJobs ) do
			Simplejob:remove( job )
			-- print( "game info: job removed" )
		end
		self.simpleJobs = {}
		
		self.selectionState = "global"
		
		self.selectedEntities = {}
		
		self.workerJobs = {}
		
		-- the mouse can leave the game area again
		-- love.mouse.setGrabbed( false )
		
		PostGame:open( self.gamestate, self.gametime )
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:debugDestroySelectedWall()
		
		-- wall selected?
		local wall = self.selectedEntity
		if not wall or wall.entityCategory ~= ENTITYCATEGORY.wall then
			return false
		end
		
		if not wall.canBeDrilled then
			print( "Game: cannot destroy solid rock!" )
			return false
		end
		
		-- remove it
		print( "Game: destroy wall at", wall.xi, wall.yi )
		self:removeWall( wall )
		-- self.selectedTile = nil
		-- self:switchToSelectionState( "tile" )
		-- self:clearEntitySelection()
		self:onEntitySelectionChanged()
		self:reselectCurrentTile()
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:debugReinforceWall( wall )
		wall.isReinforced = true
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- called from terrain
	-- e.g. when a tile was consumed by lava
	function game:onTerrainChanged( xi, yi, type )
		
		if not self:isValidTileIndex( xi, yi ) then
			return false
		end
		
		-- update blocking
		self.blocking:singleTileUpdate( xi, yi, self.terrain, self.wallData )
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:removeWall( wall )
		if not wall then
			return false
		end
		
		-- remember the position and tile index
		local x, y = wall:getPosition()
		local xi, yi = wall.xi, wall.yi
		local entityId = wall.id
		
		-- if the wall was selected, reselect the tile
		local wasSelected = wall.selected == true
		
		EntityManager:remove( wall )
		-- and be sure to update the wallData!
		self.wallData[yi][xi] = 0
		
		-- wall braking animation
		Particles:createWallBreakingParticles( x, y )
		
		-- update blocking
		self.blocking:singleTileUpdate( xi, yi, self.terrain, self.wallData )
		
		-- check up to 8 surrounding walls if they are still valid (or should collapse)
		local wallsToUpdate = {}
		for dy = -1,1 do
			for dx = -1,1 do
				local x = xi+dx
				local y = yi+dy
				if (dx ~= 0 or dy ~= 0)
				and self:isValidTileIndex( x, y ) then
					-- print( "Game: sanityCheckWall at", x, y )
					self:sanityCheckWall( x, y )
				end
			end
		end
		
		-- update wall images
		-- local walls = EntityManager:getEntitiesInCategory( ENTITYCATEGORY.wall )
		-- only need to check walls in direct vicinity
		local walls = EntityManager:getEntitiesFilter( function(e)
			return e.entityCategory == ENTITYCATEGORY.wall
					and math.abs( e.xi-xi ) < 2 and math.abs( e.yi-yi ) < 2
		end )
		self:updateWallImageCodes( walls )
		
		
		-- create debris...
		local debris = EntityManager:createEntity(
			ENTITIES.debris,
			-- also store the tile index (avoid to calculate them over and over again)
			{ x = (xi-1) * tilesize, y = (yi-1) * tilesize, xi = xi, yi = yi }
		)
		if self.fow:isTileDiscovered( xi, yi ) then
			self:addClearDebrisJob( debris )
		end
		
		-- create as many energy crystals as defined in the map data (moved into wall property)
		if wall.numberOfEnergyCrystals and wall.numberOfEnergyCrystals > 0 then
			local spread = 25
			local random = math.random
			-- for i = 1, self.energyCrystals[yi][xi] do
			for i = 1, wall.numberOfEnergyCrystals do
				local energyCrystal = EntityManager:createEntity(
					ENTITIES.energyCrystal,
					{ x = x + random(-spread,spread), y = y + random(-spread,spread) }
				)
				self:onResourceDiscovered( energyCrystal, xi, yi )
			end
		end
		
		
		-- entity selection
		-- if the wall was selected, select the tile (the debris) now
		if wasSelected then
			-- self:onEntitySelectionChanged()
			-- self:reselectCurrentTile()
			self:selectTile( xi, yi )
		end
		
		-- force exploration update
		self:updateFogOfWar()
		
		-- script callback
		self.scriptCallbacks.onWallDestroyed( entityId, xi, yi )
		
	end
	
	
	-- from x1, y1 to x2, y2 (x1 <= x2 and y1 <= y2)
	-- function game:sanityCheckWalls( x1, y1, x2, y2 )
	function game:sanityCheckWall( xi, yi )
		-- local width, height = self.mapSizeX, self.mapSizeY
		local walls = self.wallData
		
		local checkWall = function( xi, yi )
			for i, check in pairs( WallSanityChecks ) do
				local x1 = xi + check[1]
				local y1 = yi + check[2]
				local x2 = xi + check[3]
				local y2 = yi + check[4]
				local x3 = xi + check[5]
				local y3 = yi + check[6]
				if (not self:isValidTileIndex( x1, y1 ) or walls[y1][x1] > 0)
				and (not self:isValidTileIndex( x2, y2 ) or walls[y2][x2] > 0)
				and (not self:isValidTileIndex( x3, y3 ) or walls[y3][x3] > 0) then
					-- one check satisfyed = ok
					return true
				end
			end
			
			-- none of the four checks was successful
			return false
		end
		
		-- 0 = no wall, anything greater than 0 = there is a wall (type does not matter)
		if walls[yi][xi] == 0 then
			-- print( "(1) no wall at", xi, yi )
			return
		end
		
		if checkWall( xi, yi ) then
			-- wall is ok
			-- print( xi, yi, "is fine" )
			return true
		else
			-- wall must be removed
			local walls = EntityManager:getEntitiesFilter( function(e)
				return e.entityCategory == ENTITYCATEGORY.wall
					and e.xi == xi and e.yi == yi
			end )
			if walls[1] then
				-- print( xi, yi, "removed" )
				self:removeWall( walls[1] )
			else
				print( "Game:sanityCheckWall warning - no wall at", xi, yi )
			end
			-- some wall(s) must be removed
			return false
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- check if a wall has a free (no wall) neighbour tile
	function game:hasWallFreeNeighbourTile( wall )
		if not wall or wall.entityCategory ~= ENTITYCATEGORY.wall then
			return false
		end
		local xi, yi = wall.xi, wall.yi
		local hasFreeNeighbour = false
		-- x1,y1, x2,y2, ... for all 4 direct neighbours (not diagonal)
		local neighbourIndexOffsets = { 0,-1, 1,0, 0,1, -1,0 }
		for i = 1, #neighbourIndexOffsets, 2 do
			local x = xi + neighbourIndexOffsets[i]
			local y = yi + neighbourIndexOffsets[i+1]
			if self:isValidTileIndex( x, y ) then
				if self.wallData[y][x] == 0 then
					return true
				end
			end
		end
		return false
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- button callback
	function game:guiActionDrillWall()
		
		-- wall selected?
		local wall = game.selectedEntity
		if not wall or wall.entityCategory ~= ENTITYCATEGORY.wall then
			print( "Game: no wall selected" )
			return false
		end
		
		-- local xi, yi = wall.xi, wall.yi
		
		-- wall can only be destroyed if there is at least one neighbour free (no wall)
		if not self:hasWallFreeNeighbourTile( wall ) then
			print( "Game: cannot drill wall - has no empty neighbour!" )
			return false
		end
		
		-- DEBUG
		if love.keyboard.isDown( "lshift" ) then
			self:debugDestroySelectedWall()
			return false
		end
		-- /DEBUG
		
		-- if wall.entityType == ENTITIES.solidRock then
		if not wall.canBeDrilled then
			print( "Game: cannot destroy this wall!" )
			return false
		end
		
		-- add a task to a job queue
		self:addDrillWallJob( wall )
		return true
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:guiActionReinforceWall()
		
		-- wall selected?
		local wall = game.selectedEntity
		if not wall or wall.entityCategory ~= ENTITYCATEGORY.wall
		or wall.entityType == ENTITIES.solidRock then
			print( "Game: invalid or no wall selected" )
			return false
		end
		
		-- wall can only be reinforced if there is at least one neighbour free (no wall)
		if not self:hasWallFreeNeighbourTile( wall ) then
			print( "Game: cannot reinforce wall - has no empty neighbour!" )
			return false
		end
		
		-- DEBUG
		if love.keyboard.isDown( "lshift" ) then
			self:debugReinforceWall( wall )
			return false
		end
		-- /DEBUG
		
		-- already reinforced?
		if wall.isReinforced then
			return false
		end
		
		-- add a task to a job queue
		self:addReinforceWallJob( wall )
		return true
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- remove all drill wall jobs for the currently selected wall
	-- return true if there was a drill wall job that is now removed
	function game:guiActionCancelDrillWall()
		
		-- wall selected?
		local wall = game.selectedEntity
		if not wall or wall.entityCategory ~= ENTITYCATEGORY.wall
		or wall.entityType == ENTITIES.solidRock then
			print( "Game: invalid or no wall selected" )
			return false
		end
		
		local jobRemoved = false
		
		-- wall queued to be drilled?
		for i = #self.workerJobs, 1, -1 do
			local job = self.workerJobs[i]
			if job.type == "drillWall"
			and job.entityId == wall.id then
				table.remove( self.workerJobs, i )
				jobRemoved = true
				break
			end
		end
		
		-- make any worker who is currently drilling this wall stop
		local workers = EntityManager:getEntitiesFilter( function(e)
			return e.isWorker and e.isPlayerControlled and e.cancelCurrentJob
		end )
		for k, worker in pairs( workers ) do
			local job = worker.currentJob
			if job and job.type == "drillWall" and job.entityId == wall.id then
				-- make sure to not queue this job again
				job.canceled = true
				worker:cancelCurrentJob()
				break
			end
		end
		
		return jobRemoved
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- returns true if there is in general an upgrade for the passed in building entity
	function game:isBuildingUpgradeAvailable( building )
		if not building
		or not building.upgradeLevel
		or building.entityCategory ~= ENTITYCATEGORY.building then
			return false
		end
		local costs = self:getBuildingUpgradeCost( building.entityType, building.upgradeLevel+1 )
		return costs ~= nil
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- button callback
	function game:guiActionUpgradeBuilding()
		
		-- valid building selected?
		local building = self:getSelectedEntity()
		if not building
		or not building.upgradeLevel
		or building.entityCategory ~= ENTITYCATEGORY.building then
			return false
		end
		
		-- can this building be upgraded right now or is it busy?
		if not building:isReadyForUpgrade() then
			return false
		end
		-- can this building be upgraded at all?
		if not self:isBuildingUpgradeAvailable( building ) then
			return false
		end
		-- enough resources?
		local costs = self:getBuildingUpgradeCost( building.entityType, building.upgradeLevel+1 )
		if not costs or not self:hasPlayerEnoughResources( costs ) then
			return false
		end
		
		for resourceType, amount in pairs( costs ) do
			self:addToPlayersGlobalResource( resourceType, -amount )
		end
		
		-- upgrade the building (instantly)
		self:upgradeBuilding( building )
		return true
		
	end
	
	
	-- actually upgrade the building, all logic checks should have been done already
	-- TODO: particle effect
	function game:upgradeBuilding( building )
		if not building.upgrade then return false end
		
		building:upgrade()
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- vehicle or robot upgrade
	-- param: what to upgrade, for now this is a string, see ingame screen
	function game:guiActionUpgradeEntity( upgradeType )
		local entity = self:getSelectedEntity()
		if not entity
		or not entity.getPosition
		or not entity.canUpgrade
		or not entity:canUpgrade( upgradeType ) then
			return false
		end
		
		-- entity must be at a vehicle factory or robot bay (depending on entity category)
		local requiredBuildingType = ENTITIES.vehicleFactory
		if entity.entityCategory == ENTITYCATEGORY.robot then
			requiredBuildingType = ENTITIES.robotBay
		end
		local x, y = entity:getPosition()
		local xi, yi = getTileIndex( x, y )
		local building = self:getBuildingAtTile( xi, yi )
		if not building
		or building.entityType ~= requiredBuildingType then
			print( "game:guiActionUpgradeEntity - entity must be at an upgrade building!" )
			return false
		end
		
		local costs = self:getEntityUpgradeCost( upgradeType )
		if self:hasPlayerEnoughResources( costs ) then
			if entity:upgrade( upgradeType ) then
				for resourceType, amount in pairs( costs ) do
					self:addToPlayersGlobalResource( resourceType, -amount )
				end
				-- upgrade performed successfully
				print( "game:guiActionUpgradeEntity successful", upgradeType )
			end
		else
			print( "game:guiActionUpgradeEntity - resources missing!" )
		end
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- button callback (btnGameActionRepairBuilding)
	function game:guiActionRepairBuilding()
		
		-- valid building selected?
		local building = self:getSelectedEntity()
		if not building
		or not building.health
		or building.entityCategory ~= ENTITYCATEGORY.building then
			return false
		end
		
		self:addRepairBuildingJob( building )
		return true
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- button callback
	function game:guiActionTeleportRockRaiderToBuilding()
		
		-- valid building selected?
		local building = self:getSelectedEntity()
		if not building
		or building.entityCategory ~= ENTITYCATEGORY.building then
			return false
		end
		
		if building.entityType == ENTITIES.teleportPad then
			
			if not building:isReadyToTeleportUnit() then
				return
			end
			
			-- print( "Game: teleport RR to teleport pad" )
			
			-- local x, y = building:getPosition()
			local x1, y1 = building:getTeleportUnitPosition()
			-- local x2, y2 = building:getLeaveBuildingPosition()
			
			local rockRaider = EntityManager:createEntity( ENTITIES.rockRaider, { x = x1, y = y1 } )
			-- rockRaider:setLeaveBuildingState( x2, y2 )
			rockRaider:setTeleportToBuildingState( building )
			
			building:teleportUnit()
			
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- button callback
	-- teleport currently selected entity up (effectively expelling it)
	function game:guiActionExpellUnit()
		local entity = self:getSelectedEntity()
		if not entity then
			return false
		end
		
		-- cannot expell vehicle as long as it has a driver
		if entity.entityCategory == ENTITYCATEGORY.vehicle
		and entity:hasDriver() then
			return false
		end
		
		if entity.cancelCurrentJob then
			entity:cancelCurrentJob()
		end
		
		-- if the unit has a cost and is currently at a traversable place
		-- create the resources
		local x, y = entity:getPosition()
		local mini_xi, mini_yi = getMiniTileIndex( x, y )
		local costs = self:getCreateEntityCost( entity.entityType )
		if costs and not self.blocking:isMiniTileBlocked( mini_xi, mini_yi ) then
			for resourceType, amount in pairs( costs ) do
				for i = 1, amount do
					local resourceEntity = EntityManager:createEntity( resourceType, { x = x, y = y } )
					self:addCollectResourceJob( resourceEntity )
				end
			end
		end
		
		-- TODO: teleport animation
		EntityManager:remove( entity )
		self:clearEntitySelection()	-- TODO?
		-- self:onEntitySelectionChanged()
		return true
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- for any unit that can transport items (for now only for transport truck)
	-- (ugly conflicts with how to cancel the current job for Rock Raiders)
	function game:guiActionEntityDropCarriedEntities()
		local entity = self:getSelectedEntity()
		if not entity
		or not entity.dropAllCarriedEntities then
			return false
		end
		
		entity:dropAllCarriedEntities()
		return true
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:guiActionEntityCancelCurrentJob()
		local entity = self:getSelectedEntity()
		if not entity
		or not entity.cancelCurrentJob then
			return false
		end
		entity:cancelCurrentJob()
		return true
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:onBeforeEntityTeleportedUp( entity )
		if self:getSelectedEntity() == entity then
			self:clearEntitySelection()
		end
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- button callback
	function game:guiActionDestroyBuildingArea()
		local entity = self:getSelectedEntity()
		if not entity
		or entity.entityType ~= ENTITIES.buildingArea then
			return false
		end
		
		local xi, yi = entity.xi, entity.yi
		local x, y = (xi-1) * tilesize, (yi-1) * tilesize
		local debris = EntityManager:createEntity(
			ENTITIES.debris,
			{ x = x, y = y, xi = xi, yi = yi, doNotContainOre = true }
		)
		if self.fow:isTileDiscovered( xi, yi ) then
			self:addClearDebrisJob( debris )
		end
		
		-- create resource entities at this position (same as for buildings)
		self:createResourcesForBuildingToDestroy( entity )
		
		EntityManager:remove( entity )
		self:clearEntitySelection()
		return true
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- button callback
	function game:guiActionDestroyBuilding()
		local entity = self:getSelectedEntity()
		if not entity
		or entity.entityCategory ~= ENTITYCATEGORY.building then
			return false
		end
		
		self:_removeBuilding( entity )
		
		self:clearEntitySelection()
		return true
	end
	
	
	function game:createResourcesForBuildingToDestroy( entity )
		local x, y = entity:getPosition()
		local costs
		if entity.entityType == ENTITIES.constructionSite then
			-- for construction sites: create only the resources that are already there
			costs = entity:getStoredResources()
		else
			costs = self:getBuildingCost( entity.entityType )
		end
		if not costs then
			return false
		end
		-- print( "game:createResourcesForBuildingToDestroy", entity.entityType )
		for resourceType, required in pairs( costs ) do
			-- print( resourceType, required )
			for i = 1, required do
				local resourceEntity = EntityManager:createEntity(
					resourceType,
					{ x = x, y = y }
				)
				self:addCollectResourceJob( resourceEntity )
			end
		end
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- button callback
	function game:guiActionVehicleGetDriver()
		
		local vehicle = self.selectedEntity
		if not vehicle
		or vehicle.entityCategory ~= ENTITYCATEGORY.vehicle then
			return false
		end
		
		-- vehicle is already manned or somehow not able to get a driver
		if not vehicle.canBeManned or not vehicle:canBeManned() then
			return false
		end
		
		self:addVehicleGetDriverJob( vehicle )
		return true
		
	end
	
	
	-- button callback
	function game:guiActionVehicleDriverLeave()
		
		local vehicle = self.selectedEntity
		if not vehicle
		or vehicle.entityCategory ~= ENTITYCATEGORY.vehicle then
			return false
		end
		
		if not vehicle:hasDriver() then
			return false
		end
		
		-- check if the driver can safely leave the vehicle there
		-- e.g. not hovering over water or lava
		local x, y = vehicle:getPosition()
		if vehicle.getLeaveVehiclePosition then
			x, y = vehicle:getLeaveVehiclePosition()
		end
		local xi, yi = getTileIndex( x, y )
		local mini_xi, mini_yi = getMiniTileIndex( x, y )
		-- if self.terrain:getTerrainTypeAtTile( xi, yi ) ~= 1 then
		if self.blocking:isMiniTileBlocked( mini_xi, mini_yi ) then
			print( "game:guiActionVehicleDriverLeave - cannot leave here, tile blocked!" )
			return false
		end
		
		-- since this action happened from the GUI, the vehicle was selected
		local driver = vehicle:getDriver()
		
		self:vehicleDriverLeave( vehicle )
		
		-- now select the (former) driver entity
		self:selectEntity( driver )
		return true
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:vehicleDriverLeave( vehicle )
		local driver = vehicle:getDriver()
		if driver then
			driver:leaveVehicle()
		end
		self:vehicleSetDriver( vehicle, nil )
	end
	
	
	function game:vehicleSetDriver( vehicle, driver )
		vehicle:setDriver( driver )
		if driver then
			local isDriverSelected = driver.selected
			driver:setVehicle( vehicle )
			-- automatically select the vehicle if the driver was selected
			if isDriverSelected then
				self:selectEntity( vehicle )
			end
		end
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:debugDestroySelectedDebris()
		
		-- debris (tile) selected?
		local debris = self.selectedEntity
		if not debris
		or debris.entityType ~= ENTITIES.debris then
			return false
		end
		
		-- remove it
		EntityManager:remove( debris )
		
		-- self.selectedTile = nil
		-- tile remains selected!
		self:clearEntitySelection()
		self:switchToSelectionState( "tile" )
		self:reselectCurrentTile()
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- debug only: instantly create the building area
	-- at currently selected tile (if any)
	function game:debugCreateBuildingArea()
		
		if not self.selectedTile then
			return false
		end
		
		local xi, yi = self.selectedTile.xi, self.selectedTile.yi
		
		if self.terrain:getTerrainTypeAtTile( xi, yi ) ~= 1 then
			print( "Game: cannot create building area - invalid terrain!", xi, yi )
			return false
		end
		
		local entity = self:getSelectedEntity()
		
		if entity then
			print( "Game: cannot create building area, some entity selected!", xi, yi )
			print( entity.id, entity.entityType )
			return false
		end
		
		-- create building area ...
		EntityManager:createEntity(
			ENTITIES.buildingArea,
			-- also store the tile index (avoid to calculate them over and over again)
			{ x = (xi-1) * tilesize, y = (yi-1) * tilesize, xi = xi, yi = yi }
		)
		-- print( "Game: created building area at", xi, yi )
		self:reselectCurrentTile()	-- !
		
	end
	
	-- button callback
	-- at currently selected tile (if any)
	function game:guiActionCreateBuildingArea()
		
		if not self.selectedTile then
			return false
		end
		
		local xi, yi = self.selectedTile.xi, self.selectedTile.yi
		
		if self.terrain:getTerrainTypeAtTile( xi, yi ) ~= 1 then
			print( "Game: cannot create building area - invalid terrain!", xi, yi )
			return false
		end
		
		local entity = self:getSelectedEntity()
		
		if entity then
			print( "Game: cannot create building area, some entity selected!", xi, yi )
			print( entity.id, entity.entityType )
			return false
		end
		
		-- local toolStore = self:getClosestEntityOfType( ENTITIES.toolStore, xi, yi )
		-- if not toolStore then
			-- print( "Game:guiActionCreateBuildingArea - cannot create building here - no reachable toolstore." )
			-- return false
		-- end
		
		local constructionSite = self:createBuildingAreaSite( xi, yi )
		local costs = self:getBuildingCost( constructionSite.buildingType )
		if not costs then
			Logger:log( "Game:guiActionCreateBuildingArea ERROR - no costs for building type " .. constructionSite.buildingType )
			return false
		end
		
		
		for resourceType, required in pairs( costs ) do
			local available = self:getPlayerGlobalResource( resourceType )
			local drain = math.min( required, available )
			
			if toolStore then
				-- for now the position to deposit and to grab resource entities is the same
				local rx, ry = toolStore:getResourceDepositPosition()
				
				for i = 1, drain do
					local resourceEntity = EntityManager:createEntity(
						resourceType,
						{ x = rx, y = ry }
					)
					-- add a worker job for each resource
					self:addCollectResourceJob( resourceEntity )
				end
				self:addToPlayersGlobalResource( resourceType, -drain )
			end
			
			-- also create a new job to bring a resource entity of the required type to the construction site
			for i = 1, required do
				self:addBringResourceToBuildingJob( constructionSite, resourceType )
			end
		end
		
		-- select the newly create construction site
		self:selectTile( xi, yi )
		return true
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- return the building entity at the given tile or false if there is none
	function game:getBuildingAtTile( xi, yi )
		local buildings = EntityManager:getEntitiesFilter( function(e)
			return e.entityCategory == ENTITYCATEGORY.building
				and e.xi == xi and e.yi == yi
		end )
		return buildings[1] or false
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- return the building area at the given tile or false if there is none
	function game:getBuildingAreaAtTile( xi, yi )
		local buildings = EntityManager:getEntitiesFilter( function(e)
			return e.entityType == ENTITIES.buildingArea
				and e.xi == xi and e.yi == yi
		end )
		return buildings[1] or false
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:guiActionCreateBuilding( entityType )
		
		-- if not self.selectedTile then
			-- return false
		-- end
		
		-- local xi, yi = self.selectedTile.xi, self.selectedTile.yi
		
		-- buildingArea selected
		local buildingArea = self:getSelectedEntity()
		if not buildingArea
		or buildingArea.entityType ~= ENTITIES.buildingArea then
			print( "game:guiActionCreateBuilding - can only build on top of building areas!" )
			return false
		end
		local xi, yi = buildingArea.xi, buildingArea.yi
		
		if self.terrain:getTerrainTypeAtTile( xi, yi ) ~= 1 then
			-- print( "Game: cannot create building site - invalid terrain!", xi, yi )
			return false
		end
		
		-- cannot build if there is blocked!
		if self.wallData[yi][xi] > 0 then
			return false
		end
		
		-- there is already a building at this tile...
		if self:getBuildingAtTile( xi, yi ) then
			return false
		end
		
		
		-- look for the closest tool store to create (4?) construction fences
		-- if no toolstore can be reached, the building cannot be placed here!
		local toolStore = self:getClosestEntityOfType( ENTITIES.toolStore, xi, yi )
		if toolStore then
			-- create a construction site + remember which building type will be created
			-- return the entity if the creation succeded
			local constructionSite = self:createBuildingSite( entityType, xi, yi )
			if constructionSite then
				local fx, fy = toolStore:getConstructionFenceSpawnPosition()
				for index = 1, 4 do
					local fence = EntityManager:createEntity(
						ENTITIES.constructionFence,
						{ x = fx, y = fy, constructionSiteId = constructionSite.id, constructionSitePositionIndex = index, }
					)
					-- and create a new job for each of these
					self:addPlaceConstructionFenceJob( fence )
				end
			end
		else
			print( "Game info: cannot create building here - no reachable toolstore." )
		end
		
		-- moved into building code
		-- self:onBuildingCreated( building )
		
		-- success -> close new building menu
		GUImgr:setVisibility( "widGameBuildingMenu", false )
		-- select the newly create construction site
		self:selectTile( xi, yi )
		return true
		
	end
	
	
	-- create the building site for the specified (building) entity type
	function game:createBuildingSite( entityType, xi, yi )
		
		local costs = self:getBuildingCost( entityType )
		if not costs then
			Logger:log( "Game:createBuildingSite ERROR - no costs for building type " .. entityType )
			return false
		end
		local building = EntityManager:createEntity( ENTITIES.constructionSite,
			{ x = (xi-1) * tilesize, y = (yi-1) * tilesize, xi = xi, yi = yi, buildingType = entityType, costs = costs, } )
		if not building then
			Logger:log( "Game:createBuildingSite ERROR - failed to create building site!" )
			return false
		end
		-- building.isConstructionComplete = false	-- TODO: set progress to 0?
		
		-- change the appearance of the building area beneath
		local buildingArea = self:getBuildingAreaAtTile( xi, yi )
		if buildingArea then
			buildingArea:setHasBuilding( true )
		end
		
		-- self:onBuildingCreated( building )
		return building
		
	end
	
	
	-- special: create the building site for the a building area
	function game:createBuildingAreaSite( xi, yi )
		
		local entityType = ENTITIES.buildingArea
		local costs = self:getBuildingCost( entityType )
		if not costs then
			Logger:log( "Game:createBuildingAreaSite ERROR - no costs for building type " .. entityType )
			return false
		end
		local building = EntityManager:createEntity( ENTITIES.constructionSite,
			{ x = (xi-1) * tilesize, y = (yi-1) * tilesize, xi = xi, yi = yi, buildingType = entityType, costs = costs, } )
		if not building then
			print( "Game:createBuildingAreaSite WARNING - failed to create building site!" )
			return false
		end
		
		-- self:onBuildingCreated( building )
		return building
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- immediately create a building at the given tile
	-- will create a building area below if needed
	-- should only be used for mission scripting (e.g. at game start)
	function game:spawnBuilding( entityType, xi, yi )
		if not self:isValidTileIndex( xi, yi ) then
			print( "game:spawnBuilding FAILED - invalid tile index", xi, yi )
			return false
		end
		
		local x, y = (xi-1) * tilesize, (yi-1) * tilesize
		local buildingArea = self:getBuildingAreaAtTile( xi, yi )
		if buildingArea then
			buildingArea:setHasBuilding( true )
		else
			local buildingArea = EntityManager:createEntity(
				ENTITIES.buildingArea,
				{ x = x, y = y, xi = xi, yi = yi }
			)
			buildingArea:setHasBuilding( true )
		end
		
		local building = EntityManager:createEntity(
			entityType,
			{ x = x, y = y, xi = xi, yi = yi }
		)
		return building
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:getBuildingCost( entityType )
		if not COSTS.entities[entityType] then
			return nil
		end
		return Tools.copyTable( COSTS.entities[entityType] )
	end
	
	
	function game:getBuildingUpgradeCost( entityType, level )
		if not COSTS.buildingUpgrades[entityType]
		or not COSTS.buildingUpgrades[entityType][level] then
			return nil
		end
		return Tools.copyTable( COSTS.buildingUpgrades[entityType][level] )
	end
	
	
	function game:getEntityUpgradeCost( upgradeType )
		if not COSTS.entityUpgrades[upgradeType] then
			return nil
		end
		return Tools.copyTable( COSTS.entityUpgrades[upgradeType] )
	end
	
	
	function game:getCreateEntityCost( entityType )
		if not COSTS.entities[entityType] then
			return nil
		end
		return Tools.copyTable( COSTS.entities[entityType] )
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- called from RR when the fence is placed
	function game:onConstructionFencePlaced( fence )
		
		local constructionSite = EntityManager:getEntityById( fence.constructionSiteId )
		if not constructionSite then
			print( "Game:onConstructionFencePlaced error - missing construction site" )
			Logger:log( "Game:onConstructionFencePlaced ERROR - missing construction site. Fence ID: " .. fence.id )
			return false
		end
		constructionSite:onFencePlaced()
		
		local xi, yi = constructionSite.xi, constructionSite.yi
		
		-- once all 4 fences are set up, the workers can start to bring resources to the building site
		if constructionSite:areAllFencesPlaced() then
			local costs = self:getBuildingCost( constructionSite.buildingType )
			if costs then
				for resourceType, required in pairs( costs ) do
					
					local available = self:getPlayerGlobalResource( resourceType )
					local drain = math.min( required, available )
					-- self:addToPlayersGlobalResource( resourceType, -drain )
					
					-- for now the position to deposit and to grab resource entities is the same
					local toolStore = self:getClosestEntityOfType( ENTITIES.toolStore, xi, yi )
					if toolStore then
						local rx, ry = toolStore:getResourceDepositPosition()
						for i = 1, drain do
							local resourceEntity = EntityManager:createEntity(
								resourceType,
								{ x = rx, y = ry }
							)
							-- add a worker job for each resource
							self:addCollectResourceJob( resourceEntity )
						end
						-- only drain if they can be created at a toolstore!
						self:addToPlayersGlobalResource( resourceType, -drain )
					else
						print( "game:onConstructionFencePlaced - last fence placed, but no toolstore in reach!" )
						Logger:log( "Game:onConstructionFencePlaced - last fence placed, but no toolstore in reach!" )
					end
					
					-- also create a new job to bring a resource entity of the required type to the construction site
					--> regardless if a toolstore is reachable or not
					for i = 1, required do
						self:addBringResourceToBuildingJob( constructionSite, resourceType )
					end
					
				end
			else
				print( "game:onConstructionFencePlaced - last fence placed, but no costs entry!" )
				Logger:log( "Game:onConstructionFencePlaced - last fence placed, but no costs table!" )
			end
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- return the construction site that is closest to the given tile position
	-- that requires at least one more resource of the passed in entity type
	function game:getClosestConstructionSiteThatNeedsResourceType( xi, yi, resourceType )
		local entities = EntityManager:getEntitiesFilter( function( e )
			if e.entityType ~= ENTITIES.constructionSite or not e.isMissingResource
			or not e.areAllFencesPlaced then
				return false
			end
			return e:areAllFencesPlaced() and e:isMissingResource( resourceType )
		end )
		
		return self:getClosestEntityFromTable( entities, xi, yi )
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- actually looks for the entity with the shortest path
	function game:getClosestEntityOfType( entityType, xi, yi )
		local entities = EntityManager:getEntitiesFilter( function( e )
			return e.entityType == entityType
		end )
		
		return self:getClosestEntityFromTable( entities, xi, yi )
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- note: coarse ground pathing only!
	-- pass in a table of entities (that have already been filtered)
	-- return the closest entity to the given tile plus the associated path
	function game:getClosestEntityFromTable( entities, xi, yi )
		local bestPath
		local bestEntity
		for _, entity in pairs( entities ) do
			local x, y = entity:getPosition()
			local target_tx, target_ty = getTileIndex( x, y )
			local path = Pathfinding.getPath( xi, yi, target_tx, target_ty, "ground_coarse" )
			if path and (not bestPath or #path < bestPath.length) then
				bestPath = {
					path = path,
					length = #path,
				}
				bestEntity = entity
			end
		end
		return bestEntity, bestPath
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- vehicle factory menu button callback
	function game:guiActionCreateVehicle( entityType )
		local building = self.selectedEntity
		if not building
		or building.entityType ~= ENTITIES.vehicleFactory then
			print( "Game:guiActionCreateVehicle ERROR - must select a vehicle factory" )
			return false
		end
		
		local costs = self:getCreateEntityCost( entityType )
		if not costs or not self:hasPlayerEnoughResources( costs ) then
			print( "game:guiActionCreateVehicle - no costs or not enough resources!" )
			return false
		end
		
		if building:startProduction( entityType ) then
			for resourceType, amount in pairs( costs ) do
				self:addToPlayersGlobalResource( resourceType, -amount )
			end
			return true
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- robot bay menu button callback
	function game:guiActionCreateRobot( entityType )
		local building = self.selectedEntity
		if not building
		or building.entityType ~= ENTITIES.robotBay then
			print( "Game:guiActionCreateRobot ERROR - must select a robot bay" )
			return false
		end
		
		-- TODO: resource check + it takes some time to produce robots
		do
			-- local x, y = building:getPosition()
			local x1, y1 = building:getTeleportUnitPosition()
			local x2, y2 = building:getLeaveBuildingPosition()
			
			local robot = EntityManager:createEntity( entityType, { x = x1, y = y1 } )
			robot:setLeaveBuildingState( x2, y2 )
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:setDefaultWorkerJobPriorities()
		
		self.workerJobPriorities["placeConstructionFence"] = 1	-- they share the same priority level
		self.workerJobPriorities["bringResourceToBuilding"] = 1	-- they share the same priority level
		self.workerJobPriorities["vehicleGetDriver"] = 2
		self.workerJobPriorities["repairBuilding"] = 3
		self.workerJobPriorities["repairErosionDamage"] = 4
		self.workerJobPriorities["drillWall"] = 5
		self.workerJobPriorities["clearDebris"] = 6
		self.workerJobPriorities["collectEnergyCrystal"] = 7
		self.workerJobPriorities["collectOre"] = 8
		self.workerJobPriorities["reinforceWall"] = 9
		self.workerJobPriorities["collectRockSample"] = 10
		
		self:updatePriorityButtons()
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- called from ingame screen
	-- params: start and target GUI elements
	function game:onPriorityButtonDragAndDropDone( e1, e2 )
		-- print( "game:onPriorityButtonDragAndDropDone", e1.id, e2.id )
		-- print( e1.index, e2.index )
		if not e1.index or not e2.index then
			return false
		end
		
		return self:swapWorkerJobPriorities( e1.index, e2.index )
	end
	
	
	-- indices of the two priorities to swap
	function game:swapWorkerJobPriorities( p1, p2 )
		
		-- get the associated job types at these two priority indices
		-- note that it is possible that multiple different job types share the same priority!
		for jobType, prio in pairs( self.workerJobPriorities ) do
			if prio == p1 then
				self.workerJobPriorities[jobType] = p2
			elseif prio == p2 then
				self.workerJobPriorities[jobType] = p1
			end
		end
		
		-- swap them
		-- self.workerJobPriorities[prio1] = p2
		-- self.workerJobPriorities[prio2] = p1
		
		-- update the buttons
		self:updatePriorityButtons()
		
	end
	
	
	-- get the job types (table of strings) for the given priority (number)
	-- note that multiple job types may share the same priority level!
	function game:getJobTypesByPriorityLevel( level )
		local jobTypes = {}
		for jobType, prio in pairs( self.workerJobPriorities ) do
			if prio == level then
				table.insert( jobTypes, jobType )
			end
		end
		return jobTypes
	end
	
	
	-- buttons for changing the priority order
	function game:updatePriorityButtons()
		
		-- print( "game:updatePriorityButtons" )
		
		-- set images for all buttons
		-- obtain the amount of buttons implicit by checking if they exist
		local index = 1
		while true do
			local button = GUImgr:getObjectById( "btnGameJobPriority"..index )
			if not button then
				break
			end
			local jobTypes = self:getJobTypesByPriorityLevel( index )
			local jobType = jobTypes[1]
			-- print( index, jobType )
			local imageData = self:getImageDataByWorkerJob( jobType )
			button:setImageData( imageData )
			
			index = index + 1
		end
		
	end
	
	
	function game:getImageDataByWorkerJob( jobType )
		if jobType == "drillWall" then
			return Images["icon_drill_wall"]
		elseif jobType == "clearDebris" then
			return Images["icon_clear_debris"]
		elseif jobType == "collectEnergyCrystal" then
			return Images["icon_collect_energy_crystal"]
		elseif jobType == "collectOre" then
			return Images["icon_collect_ore"]
		elseif jobType == "collectRockSample" then
			return Images["icon_collect_rock_sample"]
		elseif jobType == "placeConstructionFence" or jobType == "bringResourceToBuilding" then
			return Images["icon_construct_building"]
		elseif jobType == "vehicleGetDriver" then
			return Images["icon_vehicle_get_driver"]
		elseif jobType == "reinforceWall" then
			return Images["icon_reinforce_wall"]
		elseif jobType == "repairErosionDamage" then
			return Images["icon_repair_erosion_damage"]
		elseif jobType == "repairBuilding" then
			return Images["icon_repair_building"]
		end
		return nil
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- count the amount of entities of a certain type that currently exist
	function game:getNumberOfEntitiesOfType( entityType )
		if not entityType then return 0 end
		local entities = EntityManager:getEntitiesFilter( function(e)
			return e.entityType == entityType
		end )
		return #entities
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- get the highest upgradeLevel of any entity of a certain type that currently exist
	function game:getHighestUpgradeLevelOfType( entityType )
		if not entityType then return 0 end
		local highestLevel = 0
		local entities = EntityManager:getEntitiesFilter( function(e)
			return e.entityType == entityType and entity.upgradeLevel
		end )
		for _, entity in pairs( entities ) do
			if entity.upgradeLevel > highestLevel then
				highestLevel = entity.upgradeLevel
			end
		end
		return highestLevel
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- called every update tick
	function game:updateButtons()
		
		self:updateResourceOverview()
		
		self:updateConstructionMenu()
		
		self:updateBuildingButtons()
		
		self:updateEntityButtons()
		
		self:updateTileButtons()
		
		self:updateJumpToRockRaiderButton()
		
	end
	
	-- update the values for resource overview ui elements
	function game:updateResourceOverview()
		
		local energyCrystals = self:getPlayerGlobalResource( ENTITIES.energyCrystal )
		GUImgr:getObjectById( "btnGameResourceAmountEnergyCrystal" ):setText( energyCrystals )
		
		local ore = self:getPlayerGlobalResource( ENTITIES.ore )
		GUImgr:getObjectById( "btnGameResourceAmountOre" ):setText( ore )
		
		local rockSamples = self:getPlayerGlobalResource( ENTITIES.rockSample )
		GUImgr:getObjectById( "btnGameResourceAmountRockSample" ):setText( rockSamples )
		
	end
	
	
	-- enable / disable new building buttons
	function game:updateConstructionMenu()
		
		if not GUImgr:isVisible( "widGameBuildingMenu" ) then return end
		
		-- TODO: conditions must ignore units that are currently not player controlled
		
		-- condition: how many entities of a certain type exist
		local count = function( eType )
			return self:getNumberOfEntitiesOfType( eType )
		end
		-- condition: what is the highest upgrade level an entity of the given type has
		local level = function( eType )
			return self:getHighestUpgradeLevelOfType( eType )
		end
		
		-- a tool store is required for all buildings (except buildingArea)
		if count( ENTITIES.toolStore ) == 0 then
			
			GUImgr:setEnabledFlag( "btnGameActionBuildToolStore", false )
			GUImgr:setEnabledFlag( "btnGameActionBuildTeleportPad", false )
			GUImgr:setEnabledFlag( "btnGameActionBuildVehicleFactory", false )
			GUImgr:setEnabledFlag( "btnGameActionBuildPowerStation", false )
			GUImgr:setEnabledFlag( "btnGameActionBuildRobotBay", false )
			GUImgr:setEnabledFlag( "btnGameActionBuildTeslaTower", false )
			GUImgr:setEnabledFlag( "btnGameActionBuildWaterGunTower", false )
			GUImgr:setEnabledFlag( "btnGameActionBuildSupportStation", false )
			
			return
		end
		
		GUImgr:setEnabledFlag( "btnGameActionBuildToolStore", true )
		GUImgr:setEnabledFlag( "btnGameActionBuildTeleportPad", true )
		GUImgr:setEnabledFlag( "btnGameActionBuildVehicleFactory", true )
		GUImgr:setEnabledFlag( "btnGameActionBuildRobotBay", true )
		GUImgr:setEnabledFlag( "btnGameActionBuildTeslaTower", true )
		GUImgr:setEnabledFlag( "btnGameActionBuildWaterGunTower", true )
		GUImgr:setEnabledFlag( "btnGameActionBuildSupportStation", true )
		
	end
	
	
	-- vehicle factory, robot bay etc.
	-- some buttons should be enabled/disabled depending on multiple factors:
	-- the upgradeLevel of the building, if it is powered on or not,
	-- technologies the player has researched, ...
	function game:updateBuildingButtons()
		
		local building = self:getSelectedEntity()
		if not building
		or building.entityCategory ~= ENTITYCATEGORY.building then
			return
		end
		
		local entityType = building.entityType
		local upgradeLevel = building.upgradeLevel
		
		GUImgr:setEnabledFlag( "btnGameActionDestroyBuilding", true )
		
		-- building upgrade
		if building:isReadyForUpgrade() and self:isBuildingUpgradeAvailable( building ) then
			GUImgr:setEnabledFlag( "btnGameActionUpgradeBuilding", true )
		else
			GUImgr:setEnabledFlag( "btnGameActionUpgradeBuilding", false )
		end
		
		-- repair building
		if building.health then
			GUImgr:setEnabledFlag( "btnGameActionRepairBuilding", building.health:getHealthPercent() < 100 )
		end
		
		-- teleport pad
		if entityType == ENTITIES.teleportPad then
			GUImgr:setEnabledFlag( "btnGameActionTeleportRockRaiderToBuilding", building:isReadyToTeleportUnit() )
		end
		
		-- buy vehicles
		if entityType == ENTITIES.vehicleFactory then
			
			GUImgr:setEnabledFlag( "btnGameActionCreateHoverScout", true )
			
			GUImgr:setEnabledFlag( "btnGameActionCreateDigger", upgradeLevel >= 1 )
			GUImgr:setEnabledFlag( "btnGameActionCreateTransportTruck", upgradeLevel >= 1 )
			
			GUImgr:setEnabledFlag( "btnGameActionCreateLoaderDozer", upgradeLevel >= 2 )
			GUImgr:setEnabledFlag( "btnGameActionCreateMobileLaserCutter", upgradeLevel >= 2 )
			
		end
		
		-- buy robots
		
		
	end
	
	
	function game:updateEntityButtons()
		
		local entity = self:getSelectedEntity()
		if not entity then
			return
		end
		
		local entityType = entity.entityType
		
		if entityType == ENTITIES.rockRaider then
			
			-- local btnToggleAutoJobAssignment = GUImgr:getObjectById( "btnGameActionRockRaiderToggleAutomaticJobAssignment" )
			-- if entity.disableAutomaticJobAssigment then
				-- btnToggleAutoJobAssignment:setImageData( Images["icon_gear_wheel_no"] )
			-- else
				-- btnToggleAutoJobAssignment:setImageData( Images["icon_gear_wheel_yes"] )
			-- end
			
		end
		
		-- note: right now it would be enough to update these buttons when the entity selection changed
		-- but later there might be more conditions involved in the enabled state
		--> is the vehicle/robot near a correct building? + upgrade level?
		if entity.canUpgrade then
			
			GUImgr:setEnabledFlag( "btnGameActionEntityUpgradeDrillEfficiency", entity:canUpgrade( "drill_efficiency" ) )
			GUImgr:setEnabledFlag( "btnGameActionEntityUpgradeMovementSpeed", entity:canUpgrade( "movement_speed" ) )
			GUImgr:setEnabledFlag( "btnGameActionEntityUpgradeCargoCapacity", entity:canUpgrade( "cargo_capacity" ) )
			
		end
		
		
		-- also account for a tile with debris or a building area selected
		if entityType == ENTITIES.debris then
			GUImgr:setEnabledFlag( "btnGameActionRemoveDebris", true )
			GUImgr:setEnabledFlag( "btnGameActionCreateBuildingArea", false )
			GUImgr:setEnabledFlag( "btnGameToggleBuildingMenu", false )
			GUImgr:setEnabledFlag( "btnGameActionDestroyBuilding", false )
			GUImgr:setEnabledFlag( "btnGameActionDestroyBuildingArea", false )
		else
			GUImgr:setEnabledFlag( "btnGameActionRemoveDebris", false )
			
			if entityType == ENTITIES.buildingArea then
				GUImgr:setEnabledFlag( "btnGameActionCreateBuildingArea", false )
				GUImgr:setEnabledFlag( "btnGameToggleBuildingMenu", true )
				GUImgr:setEnabledFlag( "btnGameActionDestroyBuilding", false )
				GUImgr:setEnabledFlag( "btnGameActionDestroyBuildingArea", true )
			end
		end
		
	end
	
	
	function game:updateTileButtons()
		
		if not self.selectedTile then
			return false
		end
		
		local xi, yi = self.selectedTile.xi, self.selectedTile.yi
		local terrain = self.terrain
		
		if terrain:getTerrainTypeAtTile( xi, yi ) ~= 1 then
			
			-- water or lava
			GUImgr:setEnabledFlag( "btnGameActionRepairErosionDamage", false )
			
			GUImgr:setEnabledFlag( "btnGameActionRemoveDebris", false )
			GUImgr:setEnabledFlag( "btnGameActionCreateBuildingArea", false )
			GUImgr:setEnabledFlag( "btnGameToggleBuildingMenu", false )
			GUImgr:setEnabledFlag( "btnGameActionDestroyBuilding", false )
			GUImgr:setEnabledFlag( "btnGameActionDestroyBuildingArea", false )
			
		else
			
			-- dirt
			local tile = terrain:getTile( xi, yi )
			GUImgr:setEnabledFlag( "btnGameActionRepairErosionDamage", tile.lavaResistance < terrain.erosionThreshold )
			
			GUImgr:setEnabledFlag( "btnGameActionCreateBuildingArea", true )
			
			-- this is not even called... because debris selected = no tile selected
			-- local _debris = EntityManager:getEntitiesFilter( function(e)
				-- return e.entityType == ENTITIES.debris
					-- and e.xi == xi and e.yi == yi
			-- end )
			-- GUImgr:setEnabledFlag( "btnGameActionRemoveDebris", #_debris == 1 )
			-- if #_debris == 1 then
				-- print( "debris selected" )
			-- end
			
		end
		
	end
	
	
	function game:updateJumpToRockRaiderButton()
		local rr = EntityManager:getEntitiesFilter( function(e)
			return e.entityType == ENTITIES.rockRaider
				and e.isPlayerControlled
		end )
		GUImgr:setEnabledFlag( "btnGameActionJumpToRockRaider", #rr > 0 )
		GUImgr:getObjectById( "btnGameActionJumpToRockRaider" ):setText( #rr )
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- jump to the next player controlled Rock Raider
	function game:guiActionJumpToRockRaider()
		local rr = EntityManager:getEntitiesFilter( function(e)
			return e.entityType == ENTITIES.rockRaider
				and e.isPlayerControlled
		end )
		local count = #rr
		
		if count == 0 then
			return false
		elseif count == 1 then
			local entity = rr[1]
			self:jumpToSelectEntity( entity )
			self.lastJumpToRockRaiderId = entity.id
			return entity
		end
		
		-- sort by entity ID, ascending
		table.sort( rr, function(a,b) return a.id < b.id end )
		
		-- last ID was at least as high as the current highest ID, start from lowest again
		if self.lastJumpToRockRaiderId >= rr[count].id then
			local entity = rr[1]
			self:jumpToSelectEntity( entity )
			self.lastJumpToRockRaiderId = entity.id
			return entity
		end
		
		-- otherwise, pick the RR with the next highest ID
		for i = 1, count do
			local entity = rr[i]
			if entity.id > self.lastJumpToRockRaiderId then
				self:jumpToSelectEntity( entity )
				self.lastJumpToRockRaiderId = entity.id
				return entity
			end
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- select an entity and jump to it
	function game:jumpToSelectEntity( entity )
		if not entity then return false end
		self:selectEntity( entity )
		if entity.getPosition then
			local x, y = entity:getPosition()
			Camera:setPosition( x, y )
		end
		return entity
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- TODO: add a location + allow to jump to the location of the event + button to delete it
	-- + store them in a list so the player can view the last couple of notifications with their location
	-- message can be a simple string or a table with colors and strings
	-- optional: maximum display duration
	function game:newNotificationMessage( message, duration )
		local uiElement = GUImgr:getObjectById( "novGameShortMessages" )
		if not uiElement then return end
		uiElement:newMessage( message, duration )
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- boolean: game paused state
	function game:setPaused( paused )
		self.isPaused = paused
		
		-- hide some widgets
		if self.isPaused then
			self:clearEntitySelection()
			GUImgr:setVisibility( "widGameBuildingMenu", false )
			GUImgr:setVisibility( "gameMinimap", false )
			GUImgr:setVisibility( "widGameWorkerJobPriorities", false )
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- selection (GUI) state -> moved to the game screen
	-- global = player can select any entity (nothing is currently selected)
	-- wall = a wall is selected
	-- building = a building is selected
	-- entity = at least one entity is selected
	-- tile = an (empty or with debris) tile is selected
	function game:switchToSelectionState( newState )
		
		local oldState = self.selectionState
		
		GameScreen:switchToSelectionState( newState )
		
		-- switch to new state
		self.selectionState = newState
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	--[[
	- building areas can be powered if there is an active (power on) power station on top OR if at least
	one neighbour building area is powered
	- building created event: if power station -> power the building area below (if enough energy crystals)
	- player gained a resource event: if energy crystal -> power on power stations
	- building power on status changed event: if power station -> update powered on state of building area below
	- if power on status of a building area changed -> update all neighbours
	-> only update neighbour building areas with a DIFFERENT power on status to avoid an infinite loop!
	--]]
	
	function game:onEntityCreated( entity )
		-- print( "game:onEntityCreated", GetEntityTypeName(entity.entityType) )
		
		if entity.entityType == ENTITIES.powerStation then
			local buildingArea = self:getBuildingAreaAtTile( entity.xi, entity.yi )
			if buildingArea then
				if self:getPlayerGlobalResource( ENTITIES.energyCrystal) > 0 then
					buildingArea:setPowerOnState( true )
				end
			end
		end
		
	end
	
	
	-- called right before the entity is removed
	function game:onBeforeEntityDestroyed( entity )
		
		if entity.selected then
			
			-- print( "game:onBeforeEntityDestroyed - entity was selected", GetEntityTypeName(entity.entityType) )
			
			if entity.entityType == ENTITIES.debris then
				-- self:onEntitySelectionChanged()
				-- clearEntitySelection will not work properly with multiselection, but no multi select for debris
				self:clearEntitySelection()
				
				-- the following does NOT work
				-- because at this moment the entity still exists
				-- local xi, yi = entity.xi, entity.yi
				-- print( "select", xi, yi )
				-- self:selectTile( xi, yi )
			end
			
		end
		
		-- remove blocking for this building
		if entity.xi and entity.yi and entity.getBlocking then
			self.blocking:removeEntityBlockingAtTile( entity.xi, entity.yi, entity:getBlocking() )
		end
		
	end
	
	
	-- TODO: how to disable powered building areas (and their neighbours) properly?
	-- TODO: power on neighbour building areas
	function game:onEntityPowerOnStateChanged( entity, powerState )
		
		if entity.entityType == ENTITIES.powerStation then
			local buildingArea = self:getBuildingAreaAtTile( entity.xi, entity.yi )
			if buildingArea then
				if self:getPlayerGlobalResource( ENTITIES.energyCrystal) > 0 then
					buildingArea:setPowerOnState( powerState )
				end
			end
		end
		
	end
	
	
	-- e.g. a worker brings a resource entity to a tool store
	function game:onEntityDepositedEntity( carrier, carriedEntity, deposit )
		if not deposit then
			print( "game:onEntityDepositedEntity - deposit does not exist!" )
			return
		end
		local carriedEntityCategory = carriedEntity.entityCategory
		local carriedEntityType = carriedEntity.entityType
		deposit:storeEntity( carriedEntity )
		carriedEntity:setDeposit( deposit )
		EntityManager:remove( carriedEntity )
		
		if carriedEntityCategory == ENTITYCATEGORY.resource then
			
			self:addToPlayersGlobalResource( carriedEntityType, 1 )
			
			if carriedEntityType == ENTITIES.energyCrystal
			and deposit.entityType == ENTITIES.powerStation then
				-- print( "game: power on power station" )
				deposit:setPowerOnState( true )
			end
			
		end
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- triggered from within the entity code (e.g. when they were hurt)
	function game:onEntityHealthDroppedToZero( entity )
		
		if entity.entityCategory == ENTITYCATEGORY.building then
			self:_removeBuilding( entity )
		end
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	function game:_removeBuilding( entity )
		
		if entity.entityCategory ~= ENTITYCATEGORY.building then
			return false
		end
		
		-- TODO: cancel current production or research (if isPlayerControlled)
		-- if entity.cancelCurrentJob then
			-- entity:cancelCurrentJob()
		-- end
		
		-- remove blocking
		if entity.xi and entity.yi and entity.getBlocking then
			self.blocking:removeEntityBlockingAtTile( entity.xi, entity.yi, entity:getBlocking() )
		end
		
		-- create resource entities at this position
		self:createResourcesForBuildingToDestroy( entity )
		
		-- if it is a construction site...
		-- cancel all jobs of type bringResourceToBuilding for this entity
		if entity.entityType == ENTITIES.constructionSite then
			Logger:log( "game:guiActionDestroyBuilding - about to destroy a construction site. ID:", entity.id )
			
			-- the job will be removed anyway because it is considered done
			--> no need to remove jobs from the workerJobs queue
			
			-- make any worker who is currently working on this job stop
			local workers = EntityManager:getEntitiesFilter( function(e)
				return e.isWorker and e.isPlayerControlled and e.cancelCurrentJob
			end )
			for k, worker in pairs( workers ) do
				local job = worker.currentJob
				if job and job.type == "bringResourceToBuilding" and job.entityId == entity.id then
					-- make sure to not queue this job again
					job.canceled = true
					worker:cancelCurrentJob()
					Logger:log( "Cancel bringResourceToBuilding job. Worker:", worker.id )
				end
			end
		end
		
		-- update buildingArea beneath
		local buildingArea = self:getBuildingAreaAtTile( entity.xi, entity.yi )
		if buildingArea then
			buildingArea:setHasBuilding( false )
		end
		
		-- remove building
		-- TODO: teleport animation
		EntityManager:remove( entity )
		return true
		
	end
	
	----------------------------------------------------------------------------------------------------
	
	-- not required anymore, moved back to RR code, REMOVE here
	--[[
	function game:checkDepositReachableForCarriersHoldingEntity( gametime )
		
		print( "Game:checkDepositReachableForCarriersHoldingEntity..." )
		
		local workers = EntityManager:getEntitiesFilter( function(e)
			return e.isWorker and e.isPlayerControlled and e.currentJob
			and e.task and e.task == "holdEntity"
		end )
		print( "#workers:", #workers )
		
		for k, worker in pairs( workers ) do
		
			local job = worker.currentJob
			
			-- if worker.task == "holdEntity" and job then
					
				local item = EntityManager:getEntityById( job.entityId )
				if item then
					local deposit = self:getClosestDeposit( worker, item )
					if deposit then
						local ex, ey = worker:getPosition()
						local start_xi, start_yi = getMiniTileIndex( ex, ey )
						local target_x, target_y = deposit:getResourceDepositPosition()
						local target_xi, target_yi = getMiniTileIndex( target_x, target_y )
						local path = Pathfinding.getPath( start_xi, start_yi, target_xi, target_yi, "ground_fine" )
						if path then
							worker:_setMovementDestination( target_xi, target_yi, target_x, target_y )
							worker:setMovementPath( path )
							-- print( "RR: move to deposit" )
							worker.targetDepositId = deposit.id
							worker:setTask( "moveTo" )
							print( "Game:checkDepositReachableForCarriersHoldingEntity - deposit found for worker", worker.id )
							Logger:log( "Game:checkDepositReachableForCarriersHoldingEntity - deposit found for worker", worker.id )
						end
					end
				end
			-- end
		end
		
	end
	--]]
	
	----------------------------------------------------------------------------------------------------
	
	Mainloop:add( game )
	-- GUImgr:addScreen( game )
	
	Mouse:addListener( "mousePressed", game )	-- onMousePressed
	Mouse:addListener( "wheelMoved", game )	-- make sure to have onWheelMoved function
	
	UserInput:addListener( "keyPressed", game )	-- onKeyPressed
	
	-- game:switchToSelectionState( "global" )
	
	return game
end


return Game
