local MapGenerator = {}
local random = math.random
local pi = math.pi
local Round = Tools.round
local shuffle = Tools.shuffle

-- create random maps for custom missions

-- idea:
-- separate map into chunks
-- multiple random walk iterations per chunk
-- start with all solid rock
--> tiles visited by a random walker: transform solid rock into hard rock,
--> hard into loose and finally empty (no wall) tile

local empty = 0
local looseRock = 3
local hardRock = 1
local solidRock = 2

local t_dirt = 1
local t_water = 2
local t_lava = 3

----------------------------------------------------------------------------------------------------

local function IsValidTileIndex( map, x, y )
	return x > 0 and y > 0 and x <= map.sizeX and y <= map.sizeY
end

----------------------------------------------------------------------------------------------------

-- for a given tile index check all up to 4 neighbours
-- if they have a certain terrain and wall type (either could be nil to ignore)
--[[
local function CountNeighboursOfType( map, x, y, terrainType, wallType )
	local count = 0
	for dy = -1, 1 do
		for dx = -1, 1 do
			local nx = x + dx
			local ny = y + dy
			if dx ~= 0 or dy ~= 0 and IsValidTileIndex( map, nx, ny ) then
				if (not wallType or map.walls[ny][nx] == wallType)
				and (not terrainType or map.terrain[ny][nx] == terrainType) then
					count = count + 1
				end
			end
		end
	end
	return count
end
--]]

-- ...or even more versatile:
-- pass in filter function for terrain type and wall type
--> e.g. can check how many neighbours are not water
local function CountNeighboursOfType( map, x, y, fTerrain, fWall )
	local fTerrain = fTerrain or function() return true end
	local fWall = fWall or function() return true end
	local count = 0
	for dy = -1, 1 do
		for dx = -1, 1 do
			local nx = x + dx
			local ny = y + dy
			-- all 8 neighbours
			-- if dx ~= 0 or dy ~= 0 and IsValidTileIndex( map, nx, ny ) then
			-- only 4 direct neighbours
			if math.abs(dx) + math.abs(dy) == 1 and IsValidTileIndex( map, nx, ny ) then
				if fWall( map.walls[ny][nx] )
				and fTerrain( map.terrain[ny][nx] ) then
					count = count + 1
				end
			end
		end
	end
	return count
end

----------------------------------------------------------------------------------------------------

-- same as for the map editor
-- make sure that all walls are valid (have enough neighbours)
local WallSanityChecks = {
	{-1,-1, 0,-1, -1,0},	-- top left
	{0,-1, 1,-1, 1,0},	-- top right
	{-1,0, -1,1, 0, 1},	-- bottom left
	{0,1, 1,1, 1,0}	-- bottom right
}

-- for all walls: check that it has enough neighbouring walls
-- that is, at least one in x and one in y direction
local function SanityCheckWalls( map )
	local width, height = map.sizeX, map.sizeY
	local walls = map.walls
	local bad_walls = 0
	
	local checkWall = function( xi, yi )
		for i, check in pairs( WallSanityChecks ) do
			local x1 = xi + check[1]
			local y1 = yi + check[2]
			local x2 = xi + check[3]
			local y2 = yi + check[4]
			local x3 = xi + check[5]
			local y3 = yi + check[6]
			if walls[y1][x1] > 0
			and walls[y2][x2] > 0
			and walls[y3][x3] > 0 then
				-- one check satisfyed = ok
				return true
			end
		end
		
		-- none of the four checks was successful
		return false
	end
	
	local wallsToRemove = {}
	
	-- note: assuming a border of width 1 around the entire map
	--> can skip the check for these walls (they must always exist)
	for yi = 2, height-1 do
		for xi = 2, width-1 do
			-- 0 = no wall, anything greater than 0 = there is a wall (type does not matter)
			if walls[yi][xi] > 0 then
				if not checkWall( xi, yi ) then
					bad_walls = bad_walls + 1
					-- print( "MapGenerator: bad wall at", xi, yi )
					table.insert( wallsToRemove, {xi,yi} )
				end
			end
		end
	end
	
	-- remove the bad walls
	for k, v in pairs( wallsToRemove ) do
		local xi = v[1]
		local yi = v[2]
		walls[yi][xi] = 0
	end
	
	-- if bad_walls > 0 then
		-- print( "MapGenerator: removed bad walls:", bad_walls )
	-- else
		-- print( "MapGenerator: no bad walls to remove." )
	-- end
	
end

----------------------------------------------------------------------------------------------------

-- create a traversable path from one position to another
--> must be ground + no wall
-- optional: allow to take detour (enlargen path by up to 2*2 steps)
local function CreatePath( map, x1, y1, x2, y2, allowDetour )
	-- print( "CreatePath from", x1, y1, "to", x2, y2 )
	local math = math
	local getSign = Tools.getSign
	local dx = x2-x1
	local dy = y2-y1
	local signX = getSign( dx )
	local signY = getSign( dy )
	local nSteps = math.abs(dx) + math.abs(dy)
	-- print( "min. #steps:", nSteps )
	local steps = {}
	for _ = 1, math.abs(dx) do steps[#steps+1] = {signX,0} end	-- go in x direction dx times
	for _ = 1, math.abs(dy) do steps[#steps+1] = {0,signY} end	-- go in y direction dy times
	
	-- idea: add an extra step in the opposite direction (+ one step in the same direction)
	-- to add even more variety to the paths created
	-- i.e. if the path needs 4 steps towards positive x, this would work, too:
	-- using 5 steps towards +x and 1 step in -x direction
	--> problem: this might touch (or exceed) the map borders!
	--> only 
	if allowDetour then
		if math.min(x1,x2) > 1 and math.max(x1,x2) < map.sizeX then
			steps[#steps+1] = {signX,0}
			steps[#steps+1] = {-signX,0}
		end
		if math.min(y1,y2) > 1 and math.max(y1,y2) < map.sizeY then
			steps[#steps+1] = {0,signY}
			steps[#steps+1] = {0,-signY}
		end
	end
	-- print( "actual #steps:", #steps )
	
	-- randomly permutate order of the steps
	steps = shuffle( steps )
	
	local function visit( x, y )
		-- print( x ..":".. y )
		map.walls[y][x] = empty
		map.terrain[y][x] = t_dirt
	end
	
	local x = x1
	local y = y1
	-- also visit the start tile!
	visit( x, y )
	for stepIndex = 1, #steps do
		local step = steps[stepIndex]
		x = x + step[1]
		y = y + step[2]
		visit( x, y )
	end
end

----------------------------------------------------------------------------------------------------

-- deprecated
local wallReplacer = {}
wallReplacer[0] = 0
wallReplacer[1] = 3
wallReplacer[2] = 1
wallReplacer[3] = 0

-- only visit each tile once per run!
local visited = {}

local function VisitTile( map, x, y )
	-- if map.walls[y][x] == 0 then
		-- map.walls[y][x] = EMPTY
		-- map.num_changes = map.num_changes + 1
		-- return true
	-- end
	-- return false
	
	local index = x.."_"..y
	if visited[index] then return end
	-- print( "visit", x, y )
	-- print( "replace", map.walls[y][x], "->", wallReplacer[map.walls[y][x]] )
	map.walls[y][x] = wallReplacer[map.walls[y][x]]
	visited[index] = true
end

----------------------------------------------------------------------------------------------------

local function CountEmptyNeighbours( map, x, y )
	local alive = 0
	for dy = -1, 1 do
		for dx = -1, 1 do
			local nx = x + dx
			local ny = y + dy
			if dx ~= 0 or dy ~= 0 then
				-- consider edges to be always walls (dead)
				if map.walls[ny] and map.walls[ny][nx] and map.walls[ny][nx] == empty then
					alive = alive + 1
				end
			end
		end
	end
	return alive
end

----------------------------------------------------------------------------------------------------

local function RandomWalk( map, _x, _y, max_attempts, _max_changes )
	
	-- reset visited tiles before each run
	visited = {}
	
	local w = map.sizeX
	local h = map.sizeY
	local num_tiles = w*h
	
	-- select the start position
	local start_x = _x or Round(0.01 * random(30, 70) * w)
	local start_y = _y or Round(0.01 * random(30, 70) * h)
	-- print( start_x, start_y )
	VisitTile( map, start_x, start_y )
	
	local directions = {
		{ dx = -1 },
		{ dx = 1 },
		{ dy = -1 },
		{ dy = 1 },
	}
	local changes = 0
	local attempt = 1
	local max_attempts = _max_attempts or 16
	local min_changes = math.floor( num_tiles * 0.1 )
	local max_changes = Round( _max_changes or math.floor( num_tiles * 0.1 ) )
	local border = 1	-- leave the border unchanged
	
	while changes < min_changes do
		local x, y = start_x, start_y
		repeat
			local direction = Tools.randomFromTable( directions )
			x = x + (direction.dx or 0)
			y = y + (direction.dy or 0)
			if x > border and y > border
			and x <= map.sizeX - border and y <= map.sizeY - border then
				-- only count actual changes
				if VisitTile( map, x, y ) then
					changes = changes + 1
				end
			else
				-- cancel when map border reached
				break
			end
		until
			-- map.num_changes >= max_changes
			changes >= max_changes
		
		if attempt >= max_attempts then
			break
		end
		attempt = attempt + 1
		-- print( string.format( "%d: %d/%d", attempt, map.num_changes, max_changes ) )
	end
	
	-- return map
end

----------------------------------------------------------------------------------------------------

local function CircleRandomWalk( map, num_starts )
	
	-- reset visited tiles before each run
	visited = {}
	
	local w = map.sizeX
	local h = map.sizeY
	local num_tiles = w*h
	
	local center_x = Round( 0.5 * w )
	local center_y = Round( 0.5 * h )
	local radius = Round( math.min( w, h ) * 0.3 )
	local num_starts = num_starts or 8
	local a = 2 * pi / num_starts	-- angle difference
	
	for i = 1, num_starts do
		local angle = (i-1) * a
		local x = Round( center_x + math.cos(angle) * radius )
		local y = Round( center_y + math.sin(angle) * radius )
		RandomWalk( map, x, y, 1, num_tiles * 0.01 )
		-- StupidWalk( map, x, y, math.min( w, h ) )
	end
	
end

----------------------------------------------------------------------------------------------------

local function RandomScatter( map, x1, x2, y1, y2, chance )
	
	local x1 = x1 or 1
	local x2 = x2 or map.sizeX
	local y1 = y1 or 1
	local y2 = y2 or map.sizeY
	local chanceToStartAlive = chance or 0.5
	
	for yi = y1, y2 do
		for xi = x1, x2 do
			if random() < chanceToStartAlive then
				map.walls[yi][xi] = empty
			else
				map.walls[yi][xi] = solidRock
			end
		end
	end
	
end

----------------------------------------------------------------------------------------------------

local function _cellularAutomataStep( map, x1, x2, y1, y2, _deathLimit, _birthLimit )
	
	local x1 = x1 or 1
	local x2 = x2 or map.sizeX
	local y1 = y1 or 1
	local y2 = y2 or map.sizeY
	
	local deathLimit = _deathLimit or 3
	local birthLimit = _birthLimit or 4
	local map2 = {}
	
	for y = y1, y2 do
		map2[y] = {}
		for x = x1, x2 do
			local neighbours = CountEmptyNeighbours( map, x, y )
			if map.walls[y][x] == 1 then
				-- old value: alive (empty)
				if neighbours < deathLimit then
					map2[y][x] = solidRock
				else
					map2[y][x] = empty
				end
			else
				-- old value: dead (wall)
				if neighbours > birthLimit then
					map2[y][x] = empty
				else
					map2[y][x] = solidRock
				end
			end
		end
	end
	
	-- finally, copy the contents into the map
	for y = y1, y2 do
		for x = x1, x2 do
			map.walls[y][x] = map2[y][x]
		end
	end
	
end

-- pass in min and max x and y values to only consider regions of the map
local function CellularAutomata( map, x1, x2, y1, y2, _iterations, _deathLimit, _birthLimit )
	
	local iterations = _iterations or 1
	
	local x1 = x1 or 1
	local x2 = x2 or map.sizeX
	local y1 = y1 or 1
	local y2 = y2 or map.sizeY
	
	-- initial state
	local chanceToStartAlive = 0.6	-- 0.4
	for yi = y1, y2 do
		for xi = x1, x2 do
			-- if xi == x1 or yi == y1
			-- or xi == x2 or yi == y2 then
				-- borders are always walls
				-- map.walls[yi][xi] = solidRock
			-- elseif random() < chanceToStartAlive then
				-- map.walls[yi][xi] = empty
			-- end
			if random() < chanceToStartAlive then
				map.walls[yi][xi] = empty
			else
				map.walls[yi][xi] = solidRock
			end
		end
	end
	
	for i = 1, iterations do
		_cellularAutomataStep( map, x1, x2, y1, y2, _deathLimit, _birthLimit )
	end
	
end

----------------------------------------------------------------------------------------------------

-- used to generate destructible walls on top of an existing binary (wall, empty) cave layout
--> this might open up new paths (which is ok)
local function CreateRandomWallFormation( map, type, x1, x2, y1, y2 )
	
	local type = type or looseRock
	
	local x1 = x1 or 1
	local x2 = x2 or map.sizeX
	local y1 = y1 or 1
	local y2 = y2 or map.sizeY
	
	
	-- select a random position
	local x = random( x1+1, x2-1 )
	local y = random( y1+1, y2-1 )
	
	-- create a randomly shaped blob around it and fill in a certain wall type
	--> should not overwrite the border of the map
	map.walls[y][x] = type
	local directions = {
		{ dx = -1 },
		{ dx = 1 },
		{ dy = -1 },
		{ dy = 1 },
	}
	-- number of attempts
	for i = 1, random(1,4) do
		local x = x
		local y = y
		-- how large the blob can potentially become
		for j = 1, random(5) do
			-- step in any direction and update the wall tile
			local direction = Tools.randomFromTable( directions )
			x = x + (direction.dx or 0)
			y = y + (direction.dy or 0)
			if x > x1 and x < x2 and y > y1 and y < y2 then
				map.walls[y][x] = type
			else
				break
			end
		end
	end
	
end

----------------------------------------------------------------------------------------------------

-- create up to upperLimit energy crystals in the given area (chunks)
--> it might happen that fewer resources are created!
local function DistrubuteEnergyCrystals( map, x1, x2, y1, y2, upperLimit )
	
	local x1 = x1 or 1
	local x2 = x2 or map.sizeX
	local y1 = y1 or 1
	local y2 = y2 or map.sizeY
	local upperLimit = upperLimit or 5
	local walls = map.walls
	local energyCrystals = map.energyCrystals
	
	-- count how many resources actually have been created
	local nCreated = 0
	-- maximum attempts
	local maxAttempts = 64
	local attempts = 0
	
	local function canPlace( xi, yi )
		-- limit the amount of crystals in the same spot (maximum: 2)
		if energyCrystals[yi][xi] > 2 then return false end
		-- there must be a destructible wall
		local wallType = walls[yi][xi]
		return wallType == looseRock or wallType == hardRock
	end
	
	while attempts < maxAttempts
	and nCreated < upperLimit do
		local xi = random( x1, x2 )
		local yi = random( y1, y2 )
		if canPlace( xi, yi ) then
			energyCrystals[yi][xi] = energyCrystals[yi][xi] + 1
			nCreated = nCreated + 1
		end
		attempts = attempts + 1
	end
	
	return nCreated
	
end

----------------------------------------------------------------------------------------------------

local function CreateRandomLake( map, x1, x2, y1, y2 )
	
	local x1 = x1 or 1
	local x2 = x2 or map.sizeX
	local y1 = y1 or 1
	local y2 = y2 or map.sizeY
	
	-- select a random position
	local x = random( x1+1, x2-1 )
	local y = random( y1+1, y2-1 )
	
	-- create a randomly shaped blob around it and fill in water + remove walls
	--> should not overwrite the border of the map
	map.terrain[y][x] = t_water
	map.walls[y][x] = empty
	local changedTiles = {}
	local directions = {
		{ dx = -1 },
		{ dx = 1 },
		{ dy = -1 },
		{ dy = 1 },
	}
	-- number of attempts
	for i = 1, 4 do
		local x = x
		local y = y
		-- how large the blob can potentially become
		for j = 1, random(6) do
			-- step in any direction and update the tile
			local direction = Tools.randomFromTable( directions )
			x = x + (direction.dx or 0)
			y = y + (direction.dy or 0)
			if x > x1 and x < x2 and y > y1 and y < y2 then
				-- avoid that tiles are twice in this list
				if map.terrain[y][x] ~= t_water then
					changedTiles[#changedTiles+1] = {x,y}
				end
				
				-- replace
				map.terrain[y][x] = t_water
				map.walls[y][x] = empty
			else
				break
			end
		end
	end
	
	-- check: lake should not cut off any path
	--> iterate over all changed water tiles
	--> if they have at least one neighbour with a destructible wall, replace the water with normal ground
	local fTerrain = function( type ) return true end
	local fWall = function( type ) return type == looseRock or type == hardRock end
	for i = 1, #changedTiles do
		local x = changedTiles[i][1]
		local y = changedTiles[i][2]
		if CountNeighboursOfType( map, x, y, fTerrain, fWall ) > 0 then
			-- print( "CreateRandomLake:replace water with dirt", x, y )
			map.terrain[y][x] = t_dirt
		end
	end
	
	-- remove water tiles that have no water neighbour (1x1 water pit)
	do
		local countReplaced = 0
		local fTerrain = function( type ) return type == t_water end
		for y = y1, y2 do
			for x = x1, x2 do
				if map.terrain[y][x] == t_water then
					local waterNeighbours = CountNeighboursOfType( map, x, y, fTerrain, nil )
					if waterNeighbours == 0 then
						-- print( "CreateRandomLake:replace 1x1 water with dirt", x, y )
						countReplaced = countReplaced + 1
						map.terrain[y][x] = t_dirt
					end
				end
			end
		end
		-- print( "replaced 1x1 water:", countReplaced )
	end
	
end

----------------------------------------------------------------------------------------------------

local function CreateRandomRiver( map, x1, x2, y1, y2, maxLength, width )
	
	local width = width or 1
	local maxLength = maxLength or map.sizeX + map.sizeY
	
	local x1 = x1 or 1
	local x2 = x2 or map.sizeX
	local y1 = y1 or 1
	local y2 = y2 or map.sizeY
	
	-- select the start position
	-- either start from the top or left side
	local directions = RandomBag:create()
	local start_x = 1
	local start_y = 1
	
	if random() < 0.5 then
		if random() < 0.5 then
			-- start: x=1 - top
			--> prefer to extend down (+x)
			directions:add( {dx=1}, 50 )
			directions:add( {dx=-1}, 10 )
			directions:add( {dy=1}, 20 )
			directions:add( {dy=-1}, 20 )
			start_x = x1
		else
			-- start: x=max - bottom
			--> prefer to extend up (-x)
			directions:add( {dx=-1}, 50 )
			directions:add( {dx=1}, 10 )
			directions:add( {dy=-1}, 20 )
			directions:add( {dy=1}, 20 )
			start_x = x2
		end
		start_y = Round( y1+1, y2-1 )
	else
		if random() < 0.5 then
			-- start: y=1 - left
			--> prefer to extend right (+y)
			directions:add( {dy=1}, 50 )
			directions:add( {dy=-1}, 10 )
			directions:add( {dx=1}, 20 )
			directions:add( {dx=-1}, 20 )
			start_y = y1
		else
			-- start: y=max - right
			--> prefer to extend left (-y)
			directions:add( {dy=-1}, 50 )
			directions:add( {dy=1}, 10 )
			directions:add( {dx=1}, 20 )
			directions:add( {dx=-1}, 20 )
			start_y = y2
		end
		start_x = Round( x1+1, x2-1 )
	end
	
	local length = 0
	local x, y = start_x, start_y
	local changedTiles = {}
	
	while true do
		length = length + 1
		local direction = directions:getRandom()
		x = x + (direction.dx or 0)
		y = y + (direction.dy or 0)
		
		if x > x1 and x < x2 and y > y1 and y < y2 then
			-- avoid that tiles are twice in this list
			-- if map.terrain[y][x] ~= t_water then
				-- changedTiles[#changedTiles+1] = {x,y}
			-- end
			
			-- replace
			map.terrain[y][x] = t_water
			map.walls[y][x] = empty
		end
		
		-- TODO: if width > 1 then also cover the neighbour tile(s) in the proper direction
		--> if start_x == 1 then also change y+1
		
		if x == map.sizeX or y == map.sizeY or length > maxLength then
			break
		end
	end
	
	-- check: river should not cut off any path
	-- other than for lakes, this does not feel good for rivers
	--> iterate over all changed water tiles
	--> if they have at least one neighbour with a destructible wall, replace the water with normal ground
	local fTerrain = function( type ) return true end
	local fWall = function( type ) return type == looseRock or type == hardRock end
	for i = 1, #changedTiles do
		local x = changedTiles[i][1]
		local y = changedTiles[i][2]
		if CountNeighboursOfType( map, x, y, fTerrain, fWall ) > 0 then
			-- print( "CreateRandomRiver:replace water with dirt", x, y )
			map.terrain[y][x] = t_dirt
		end
	end
	
end

----------------------------------------------------------------------------------------------------

function MapGenerator:create()
	local generator = { sizeX = 0, sizeY = 0, nWaterTiles = 0, nLavaTiles = 0, }
	
	function generator:createMap( sizeX, sizeY, options )
		
		local options = options or {}
		local seed = options.seed or os.time()
		print( "MapGenerator seed:", seed )
		math.randomseed( seed )
		
		local map = {}
		map.sizeX = sizeX or 32
		map.sizeY = sizeY or 32
		
		map.terrain = {}
		map.walls = {}
		map.energyCrystals = {}
		
		-- prepare the map
		for y = 1, map.sizeY do
			map.terrain[y] = {}
			map.walls[y] = {}
			map.energyCrystals[y] = {}
			for x = 1, map.sizeX do
				-- normal ground
				map.terrain[y][x] = t_dirt
				-- fill everything with solid rock first
				map.walls[y][x] = solidRock
				map.energyCrystals[y][x] = 0
			end
		end
		
		
		-- partition the entire cave area into smaller "rooms"
		-- only divide a room again if it is larger than this in borh directions:
		local minRoomSize = 6
		local function partition( room )
			-- print( "partition room..." )
			-- print( "edges:", room.x1, room.x2, room.y1, room.y2 )
			-- print( "size:", room.x2-room.x1, "x", room.y2-room.y1 )
			local roomA, roomB
			-- check size condition
			local rx = room.x2-room.x1
			local ry = room.y2-room.y1
			if rx > minRoomSize and ry > minRoomSize then
				-- choose partition direction
				local divX, divY
				if rx > ry then
					-- divide into left and right
					divX = room.x1 + Round( random(4,6) * 0.1 * rx )
				elseif rx < ry then
					-- divide into top and bottom
					divY = room.y1 + Round( random(4,6) * 0.1 * ry )
				else	-- sizes are equal, choose randomly
					if random() < 0.5 then
						divX = room.x1 + Round( random(4,6) * 0.1 * rx )
					else
						divY = room.y1 + Round( random(4,6) * 0.1 * ry )
					end
				end
				
				if divX then
					-- print( "divX:", divX )
					roomA = {x1=room.x1, x2=divX, y1=room.y1, y2=room.y2}
					roomB = {x1=divX+1, x2=room.x2, y1=room.y1, y2=room.y2}
				elseif divY then
					-- print( "divY:", divY )
					roomA = {x1=room.x1, x2=room.x2, y1=room.y1, y2=divY}
					roomB = {x1=room.x1, x2=room.x2, y1=divY+1, y2=room.y2}
				end
				
				return roomA, roomB
			end
			
			-- this room was not partitioned further
			-- print( "do not divide further, too small" )
			return false
			-- break
		end
		
		-- this could probably be done more elegantly using binary trees...
		local rooms = {}
		local neighbours = {}
		-- first room = entire cave
		rooms[1] = {x1=1,x2=map.sizeX, y1=1,y2=map.sizeY}
		local i = 0
		
		while i < #rooms do
			i = i+1
			-- print( "partition", i, "..." )
			local A, B = partition( rooms[i] )
			if A then
				rooms[#rooms+1] = A
				rooms[#rooms+1] = B
				neighbours[#rooms] = #rooms-1
				-- neighbours[#rooms-1] = #rooms
			else
				rooms[i].isLeaf = true
				-- print( i, "is a final room" )
			end
		end
		
		-- rooms inside partitions might be smaller than the partition (that means more surrounding walls)
		for i = 1, #rooms do
			local room = rooms[i]
			if room.isLeaf then
				if room.x2-room.x1 > 4 then
					room.x1 = room.x1 + random(0,1)
					room.x2 = room.x2 - random(0,1)
				end
				if room.y2-room.y1 > 4 then
					room.y1 = room.y1 + random(0,1)
					room.y2 = room.y2 - random(0,1)
				end
			end
		end
		
		-- print( "#rooms:", #rooms )
		-- edit each room individually
		-- TODO: put all the random walker etc. in here
		--> might use different methods per room!
		--> need to change the parameter to only edit parts of the whole map (=rooms)
		-- problem: start/end tile of paths that will be created might not actually connect empty caves
		--> sometimes caves might be unreachable!
		local nLeaves = 0
		-- ignore the first room (implicit, because it is usually not a leaf)
		for i = 1, #rooms do
			local room = rooms[i]
			if room.isLeaf then
				nLeaves = nLeaves + 1
				-- print( "room size:", room.x2-room.x1, "x", room.y2-room.y1 )
				
				-- for now make all rooms empty, surrounded by walls
				-- for y = room.y1+1, room.y2-1 do
					-- for x = room.x1+1, room.x2-1 do
						-- map.walls[y][x] = empty
					-- end
				-- end
				
				local iterations = 2
				CellularAutomata( map, room.x1, room.x2, room.y1, room.y2, iterations )
				-- RandomScatter( map, room.x1, room.x2, room.y1, room.y2, 0.5 )
				
			end
		end
		-- print( "#final rooms:", nLeaves )
		
		-- create paths (remove walls) to connect all neighbouring rooms
		for k,v in pairs(neighbours) do
			-- print( "connect room", k, "and", v )
			local A = rooms[k]
			local B = rooms[v]
			-- select a random position from each room
			-- and create a connection (remove walls in between)
			local x1 = random( A.x1+1, A.x2-1 )
			local x2 = random( B.x1+1, B.x2-1 )
			local y1 = random( A.y1+1, A.y2-1 )
			local y2 = random( B.y1+1, B.y2-1 )
			-- print( "create path from", x1, y1, "to", x2, y2 )
			CreatePath( map, x1, y1, x2, y2, true )
		end
		--
		
		-- enforce wall border
		--> only required if the borders might be touched by room-filling code
		for x = 1, map.sizeX do
			map.walls[1][x] = solidRock
			map.walls[map.sizeY][x] = solidRock
		end
		for y = 1, map.sizeY do
			map.walls[y][1] = solidRock
			map.walls[y][map.sizeX] = solidRock
		end
		
		-- randomly distribute blobs of various shapes of hard and loose rock
		for _ = 1, random( map.sizeX, map.sizeX + map.sizeY ) do
			local type = random() < 0.5 and looseRock or hardRock
			CreateRandomWallFormation( map, type )
		end
		
		-- if water should be generated, select one or more locations at the edge
		-- and perform a random walk (low chance to change direction) until border reached
		--> remove all walls, set terrain to water
		-- problem: this might cut off some areas, which is really not OK
		--> idea: if destructible wall next to water - add one tile of ground (empty)
		--> this works great for lakes, but looks bad for rivers
		if options.water and options.water > 1 then
			local minSize = math.min( map.sizeX, map.sizeY )
			
			-- river
			if options.water == 2 or options.water == 4 then
				for i = 1, math.max( 1, Round(minSize/16) ) do
					-- print( "create river" )
					CreateRandomRiver( map )
				end
			end
			
			-- lake
			if options.water == 3 or options.water == 4 then
				for i = 1, math.max( 1, Round(minSize/16) ) do
					-- print( "create lake" )
					CreateRandomLake( map )
				end
			end
		end
		
		-- force an empty area (+normal ground) to create a small player base
		--> this tile index is also needed inside the script to create initial buildings (or units)
		-- TODO: create the base inside a cave + extend if needed
		--> or if the base is surrounded by solidRock (or water?) create another tunnel to a random cave
		local baseTileX = Round( random( 3, 7 ) * 0.1 * map.sizeX )
		local baseTileY = Round( random( 3, 7 ) * 0.1 * map.sizeY )
		map.walls[baseTileY][baseTileX] = empty
		map.walls[baseTileY][baseTileX+1] = empty
		map.walls[baseTileY+1][baseTileX] = empty
		map.walls[baseTileY+1][baseTileX+1] = empty
		map.terrain[baseTileY][baseTileX] = t_dirt
		map.terrain[baseTileY][baseTileX+1] = t_dirt
		map.terrain[baseTileY+1][baseTileX] = t_dirt
		map.terrain[baseTileY+1][baseTileX+1] = t_dirt
		-- print( "base position:", baseTileX, baseTileY )
		local neighbours = {
			-- left
			{baseTileX-1, baseTileY},
			{baseTileX-1, baseTileY+1},
			-- right
			{baseTileX+2, baseTileY},
			{baseTileX+2, baseTileY+1},
			-- top
			{baseTileX, baseTileY-1},
			{baseTileX+1, baseTileY-1},
			-- bottom
			{baseTileX, baseTileY+2},
			{baseTileX+1, baseTileY+2},
		}
		local baseIsEnclosed = true
		for i = 1, #neighbours do
			local n = neighbours[i]
			local nx, ny = n[1], n[2]
			if map.walls[ny][nx] ~= solidRock
			and map.terrain[ny][nx] == t_dirt then
				baseIsEnclosed = false
				break
			end
		end
		if baseIsEnclosed then
			-- where should this path lead?
			--> just select one random (leaf) room and it's first corner
			--> careful: the room might start at x=1 or y=1
			local randomRoom
			for i = 1, #rooms do
				local room = rooms[i]
				if room.isLeaf then
					randomRoom = room
					break
				end
			end
			-- print( "info: base is enclosed!" )
			local x2 = random(randomRoom.x1 + 1, randomRoom.x2 - 1)
			local y2 = random(randomRoom.y1 + 1, randomRoom.y2 - 1)
			-- print( "create path from", baseTileX, baseTileY, "to",  x2, y2 )
			CreatePath( map, baseTileX, baseTileY, x2, y2, true )
		end
		
		SanityCheckWalls( map )
		
		-- distribute some energy crystals...
		-- idea: divide te map into small cunks (e.g. 8x8) and set a target value per chunk
		-- assume the map size is a multiple of 8 in both dimensions
		local chunkSize = 8
		local nChunksX = map.sizeX / chunkSize
		local nChunksY = map.sizeY / chunkSize
		local upperLimit = 5	-- do not create more than X energy crystals per chunk
		local minimumTotalEnergyCrystals = map.sizeX	-- TODO: base value?
		if options.energyCrystalAmount then
			if options.energyCrystalAmount == "few" then
				upperLimit = 3
				minimumTotalEnergyCrystals = Round( minimumTotalEnergyCrystals * 0.6 )
			elseif options.energyCrystalAmount == "normal" then
				upperLimit = 4
				-- minimumTotalEnergyCrystals = Round( minimumTotalEnergyCrystals * 0.6 )
			elseif options.energyCrystalAmount == "plenty" then
				upperLimit = 5
				minimumTotalEnergyCrystals = Round( minimumTotalEnergyCrystals * 1.2 )
			end
		end
		local totalEnergyCrystalsCreated = 0
		for yi = 1, nChunksY do
			for xi = 1, nChunksX do
				local x1 = 1 + (xi-1) * chunkSize
				local x2 = xi * chunkSize - 1
				local y1 = 1 + (yi-1) * chunkSize
				local y2 = yi * chunkSize - 1
				local nCreated = DistrubuteEnergyCrystals( map, x1, x2, y1, y2, upperLimit )
				totalEnergyCrystalsCreated = totalEnergyCrystalsCreated + nCreated
			end
		end
		-- print( "created energy crystals:", totalEnergyCrystalsCreated .. "/" .. minimumTotalEnergyCrystals, "(have/minimum)" )
		-- TODO: if the amount of energy crystals is below a minimum threshold,
		-- perform a few iterations on some chunks again
		
		local script = {}
		script.onGameStart = function()
			print( "Custom mission started." )
			
			-- local x, y = tileIndexToPosition( 8, 8 )
			-- EntityManager:createEntity( ENTITIES.rockRaider, { x = x, y = y } )
			local building = Game:spawnBuilding( ENTITIES.teleportPad, baseTileX, baseTileY )
			local x, y = building:getPosition()
			Camera:setPosition( x, y )
			Game:spawnBuilding( ENTITIES.toolStore, baseTileX, baseTileY+1 )
			-- building.explorationRange = 7	-- increase exploration range for easier testing
			--> this is really bad! causes lags or even crashes when too large!
			
			-- Game.fow.debugExploreAll = true
			
		end
		
		-- return a "map" and a script
		return map, script
		
	end
	
	
	return generator
end

return MapGenerator
